# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import autoai_libs.transformers.exportable
import numpy as np

import lale.docstrings
import lale.operators


class _CatImputerImpl:
    def __init__(self, *args, **kwargs):
        self._wrapped_model = autoai_libs.transformers.exportable.CatImputer(
            *args, **kwargs
        )

    def fit(self, X, y=None, **fit_params):
        self._wrapped_model.fit(X, y, **fit_params)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "strategy",
                "missing_values",
                "sklearn_version_family",
                "activate_flag",
            ],
            "relevantToOptimizer": ["strategy"],
            "properties": {
                "strategy": {
                    "description": "The imputation strategy.",
                    "anyOf": [
                        {
                            "enum": ["mean"],
                            "description": "Replace using the mean along each column. Can only be used with numeric data.",
                        },
                        {
                            "enum": ["median"],
                            "description": "Replace using the median along each column. Can only be used with numeric data.",
                        },
                        {
                            "enum": ["most_frequent"],
                            "description": "Replace using most frequent value each column. Used with strings or numeric data.",
                        },
                        {
                            "enum": ["constant"],
                            "description": "Replace with fill_value. Can be used with strings or numeric data.",
                        },
                    ],
                    "default": "mean",
                },
                "missing_values": {
                    "description": "The placeholder for the missing values. All occurrences of missing_values will be imputed.",
                    "anyOf": [
                        {"type": "number"},
                        {"type": "string"},
                        {"enum": [np.nan]},
                        {"enum": [None]},
                    ],
                    "default": np.nan,
                },
                "sklearn_version_family": {
                    "description": "The sklearn version for backward compatibiity with versions 019 and 020dev. Currently unused.",
                    "enum": ["20", "21", "22", "23", "24", None, "1"],
                    "default": None,
                },
                "activate_flag": {
                    "description": "If False, transform(X) outputs the input numpy array X unmodified.",
                    "type": "boolean",
                    "default": True,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        },
        "y": {"laleType": "Any"},
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        }
    },
}

_output_transform_schema = {
    "description": "Features; the outer array is over samples.",
    "anyOf": [
        {"type": "array", "items": {"laleType": "Any"}},
        {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}},
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Operator from `autoai_libs`_. Missing value imputation for categorical features, currently internally uses the sklearn SimpleImputer_.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs
.. _SimpleImputer: https://scikit-learn.org/0.20/modules/generated/sklearn.impute.SimpleImputer.html#sklearn-impute-simpleimputer""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_libs.cat_imputer.html",
    "import_from": "autoai_libs.transformers.exportable",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

CatImputer = lale.operators.make_operator(_CatImputerImpl, _combined_schemas)

autoai_libs_version_str = getattr(autoai_libs, "__version__", None)
if autoai_libs_version_str is not None:
    import typing

    from packaging import version

    from lale.schemas import AnyOf, Array, Enum, Float, Not, Null, Object, String

    autoai_libs_version = version.parse(autoai_libs_version_str)

    if autoai_libs_version >= version.Version("1.12.18"):
        CatImputer = typing.cast(
            lale.operators.PlannedIndividualOp,
            CatImputer.customize_schema(
                set_as_available=True,
                constraint=[
                    AnyOf(
                        desc="fill_value and fill_values cannot both be specified",
                        forOptimizer=False,
                        types=[Object(fill_value=Null()), Object(fill_values=Null())],
                    ),
                    AnyOf(
                        desc="if strategy=constants, the fill_values cannot be None",
                        forOptimizer=False,
                        types=[
                            Object(strategy=Not(Enum(["constants"]))),
                            Not(Object(fill_values=Null())),
                        ],
                    ),
                ],
                fill_value=AnyOf(
                    types=[Float(), String(), Enum(values=[np.nan]), Null()],
                    desc="The placeholder for fill value used in constant strategy",
                    default=None,
                ),
                fill_values=AnyOf(
                    types=[
                        Array(
                            items=AnyOf(
                                types=[Float(), String(), Enum(values=[np.nan]), Null()]
                            )
                        ),
                        Null(),
                    ],
                    desc="The placeholder for fill values used in constants strategy",
                    default=None,
                ),
                sklearn_version_family=Enum(
                    desc="The sklearn version for backward compatibiity with versions 019 and 020dev. Currently unused.",
                    values=["20", "21", "22", "23", "24", None, "1"],
                    default=None,
                ),
                strategy=AnyOf(
                    types=[
                        Enum(
                            values=["most_frequent"],
                            desc="Replace using most frequent value each column. Used with strings or numeric data.",
                        ),
                        Enum(
                            values=["constant"],
                            desc="Replace with fill_value. Can be used with strings or numeric data.",
                        ),
                        Enum(
                            values=["constants"],
                            desc="Replace missing values in columns with values in fill_values list. Can be used with list of strings or numeric data.",
                        ),
                    ],
                    desc="The imputation strategy.",
                    default="most_frequent",
                ),
            ),
        )

lale.docstrings.set_docstrings(CatImputer)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from autoai_ts_libs.sklearn.mvp_windowed_transformed_target_estimators import (  # type: ignore # noqa
    AutoaiWindowTransformedTargetRegressor as model_to_be_wrapped,
)
from sklearn.pipeline import Pipeline, make_pipeline

import lale.docstrings
import lale.operators


class _AutoaiWindowTransformedTargetRegressorImpl:
    def __init__(
        self,
        feature_columns=None,
        target_columns=None,
        regressor=None,
        lookback_window=10,
        prediction_horizon=1,
        scaling_func=None,
        inverse_scaling_func=None,
        check_inverse=False,
        short_name="",
        one_shot=False,
        row_mean_center=False,
        estimator_prediction_type="forecast",
        time_column=-1,
        random_state=42,
    ):
        if regressor is None:
            nested_op = None
        elif isinstance(regressor, lale.operators.TrainableIndividualOp):
            nested_op = make_pipeline(regressor.impl)
        elif isinstance(regressor, lale.operators.BasePipeline):
            nested_op = regressor.export_to_sklearn_pipeline()
        elif isinstance(regressor, Pipeline):
            nested_op = regressor
        else:
            # TODO: What is the best way to handle this case?
            nested_op = None
        self._hyperparams = {
            "feature_columns": feature_columns,
            "target_columns": target_columns,
            "regressor": nested_op,
            "lookback_window": lookback_window,
            "prediction_horizon": prediction_horizon,
            "scaling_func": scaling_func,
            "inverse_scaling_func": inverse_scaling_func,
            "check_inverse": check_inverse,
            "short_name": short_name,
            "one_shot": one_shot,
            "row_mean_center": row_mean_center,
            "estimator_prediction_type": estimator_prediction_type,
            "time_column": time_column,
            "random_state": random_state,
        }
        self._wrapped_model = model_to_be_wrapped(**self._hyperparams)

    def fit(self, X, y):
        self._wrapped_model.fit(X, y)
        return self

    def predict(self, X=None, prediction_type=None, **predict_params):
        return self._wrapped_model.predict(X, prediction_type, **predict_params)


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "feature_columns",
                "target_columns",
                "regressor",
                "lookback_window",
                "prediction_horizon",
                "scaling_func",
                "inverse_scaling_func",
                "check_inverse",
                "short_name",
                "one_shot",
                "row_mean_center",
                "estimator_prediction_type",
                "time_column",
                "random_state",
            ],
            "relevantToOptimizer": ["lookback_window"],
            "properties": {
                "feature_columns": {
                    "description": "",  # TODO: document and refine type
                    "laleType": "Any",
                    "default": None,
                },
                "target_columns": {
                    "description": "",  # TODO: document and refine type
                    "laleType": "Any",
                    "default": None,
                },
                "regressor": {
                    "description": """Regressor object.
This regressor will automatically be cloned each time prior to fitting.
If regressor is None, LinearRegression() is created and used.""",
                    "anyOf": [{"laleType": "operator"}, {"enum": [None]}],
                    "default": None,
                },
                "lookback_window": {
                    "description": "The number of time points in the window of data to use as predictors in the estimator.",
                    "type": "integer",
                    "default": 10,
                },
                "prediction_horizon": {
                    "description": "The number of time points to predict into the future. The estimator(s) will be trained to predict all of these time points.",
                    "type": "integer",
                    "default": 1,
                },
                "scaling_func": {
                    "description": """(deprecated) Function to apply to y before passing to fit.
The function needs to return a 2-dimensional array.
If func is None, the function used will be the identity function.""",
                    "laleType": "Any",
                    "default": None,
                },
                "inverse_scaling_func": {
                    "description": """(deprecated) Function to apply to the prediction of the regressor.
The function needs to return a 2-dimensional array.
The inverse function is used to return predictions to the same space of the original training labels.""",
                    "laleType": "Any",
                    "default": None,
                },
                "check_inverse": {
                    "description": """Whether to check that transform followed by inverse_transform or func followed by inverse_func leads to the original targets.""",
                    "type": "boolean",
                    "default": False,
                },
                "short_name": {
                    "description": "Short name to be used for this estimator.",
                    "type": "string",
                    "default": "",
                },
                "one_shot": {
                    "description": "(deprecated)",
                    "anyOf": [{"type": "boolean"}, {"enum": [None]}],
                    "default": False,
                },
                "row_mean_center": {
                    "description": "Whether to apply the row mean center transformation to the data. If true, windows of the data according to lookback_window aree created and then those rows are normalized.",
                    "type": "boolean",
                    "default": False,
                },
                "estimator_prediction_type": {
                    "description": "Defines what predictions are returned by the predict functionality. Forecast: only generate predictions for the out of sample data, i.e., for the prediction window immediately following the input data. Rowwise: make predictions for each time point in the input data.",
                    "enum": [
                        "forecast",
                        "rowwise",
                    ],
                    "default": "forecast",
                },
                "time_column": {
                    "description": "",  # TODO: document and refine type
                    "laleType": "Any",
                    "default": -1,
                },
                "random_state": {
                    "description": "",  # TODO: document and refine type
                    "laleType": "Any",
                    "default": 42,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        },
        "y": {"laleType": "Any"},
    },
}

_input_predict_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {  # Handles 1-D arrays as well
            "anyOf": [
                {"type": "array", "items": {"laleType": "Any"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"laleType": "Any"}},
                },
            ]
        },
        "prediction_type": {
            "enum": [
                "forecast",
                "rowwise",
            ]
        },
    },
}

_output_predict_schema = {
    "description": "Features; the outer array is over samples.",
    "anyOf": [
        {"type": "array", "items": {"laleType": "Any"}},
        {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}},
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Operator from `autoai_ts_libs`_.

.. _`autoai_ts_libs`: https://pypi.org/project/autoai-ts-libs""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_ts_libs.autoai_window_transformed_target_regressor.html",
    "import_from": "autoai_ts_libs.sklearn.mvp_windowed_transformed_target_estimators",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}

AutoaiWindowTransformedTargetRegressor = lale.operators.make_operator(
    _AutoaiWindowTransformedTargetRegressorImpl, _combined_schemas
)

lale.docstrings.set_docstrings(AutoaiWindowTransformedTargetRegressor)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# A shim for compatibility across 3.7.
# pre 3.7, we need to inherit from the GenericMeta class (which inherits from ABCmeta)
# to use Generic (which we want to do)
# post 3.7, GenericMeta no longer exists
import sys
from abc import ABCMeta


class VisitorMeta(type):
    """This meta class adds a private _accept method that calls visitCLASSNAME on the visitor.
    It does not currently support inheritance: you need to define the visitC method for subclasses
    explicitly.
    The private _accept method should be called via the Visitor#acccept method
    """

    def __init__(cls, *args, **kwargs):
        super(VisitorMeta, cls).__init__(*args, **kwargs)
        selector = """
        from lale.util import VisitorPathError
        try:
            return visitor.visit{}(self, *args, **kwargs)
        except VisitorPathError as e:
            e.push_parent_path(self)
            raise
        except BaseException as e:
            raise VisitorPathError([self]) from e
        """.format(
            getattr(cls, "__name__", "???")
        )
        _accept_code = "def _accept(self, visitor, *args, **kwargs):\n\t{}".format(
            selector
        )
        ll = {}
        exec(_accept_code, globals(), ll)
        setattr(cls, "_accept", ll["_accept"])


if sys.version_info < (3, 7, 0):
    from typing import GenericMeta  # type: ignore
else:
    global GenericMeta
    GenericMeta = ABCMeta  # type: ignore


class AbstractVisitorMeta(VisitorMeta, GenericMeta):
    """This meta class adds an _accept method that calls visitCLASSNAME on the visitor.
    It does not currently support inheritance: you need to define the visitC method for subclasses
    explicitly.
    The private _accept method should be called via the Visitor#acccept method.
    """

    def __init__(cls, *args, **kwargs):
        super(AbstractVisitorMeta, cls).__init__(*args, **kwargs)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.preprocessing

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Generate polynomial and interaction features.",
    "allOf": [
        {
            "type": "object",
            "required": ["include_bias"],
            "relevantToOptimizer": ["degree", "interaction_only", "include_bias"],
            "additionalProperties": False,
            "properties": {
                "degree": {
                    "type": "integer",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 3,
                    "default": 2,
                    "description": "The degree of the polynomial features.",
                },
                "interaction_only": {
                    "type": "boolean",
                    "default": False,
                    "description": "If true, only interaction features are produced: features that are products of at most degree distinct input features (so not x[1] ** 2, x[0] * x[2] ** 3, etc.).",
                },
                "include_bias": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True (default), then include a bias column, the feature in which all polynomial powers are zero (i.e. a column of ones - acts as an intercept term in a linear model).",
                },
            },
        }
    ],
}

_input_fit_schema = {
    "description": "Compute number of output features.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The data.",
        },
        "y": {},
    },
}
_input_transform_schema = {
    "description": "Transform data to polynomial features",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The data to transform, row by row.",
        },
    },
}
_output_transform_schema = {
    "description": "The matrix of features, where NP is the number of polynomial",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Polynomial features`_ transformer from scikit-learn.

.. _`Polynomial features`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PolynomialFeatures.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.polynomial_features.html",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

PolynomialFeatures: lale.operators.PlannedIndividualOp
PolynomialFeatures = lale.operators.make_operator(
    sklearn.preprocessing.PolynomialFeatures, _combined_schemas
)

if sklearn.__version__ >= "0.21":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.PolynomialFeatures.html
    # new: https://scikit-learn.org/0.23/modules/generated/sklearn.preprocessing.PolynomialFeatures.html
    from lale.schemas import Enum

    PolynomialFeatures = PolynomialFeatures.customize_schema(
        order=Enum(
            values=["C", "F"],
            desc="Order of output array in the dense case. 'F' order is faster to compute, but may slow down subsequent estimators.",
            default="C",
        ),
        set_as_available=True,
    )


lale.docstrings.set_docstrings(PolynomialFeatures)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings
from typing import Any

# This method (and the to_lale() method on the returned value)
# are the only ones intended to be exported

# This entire file is deprecated, and will be removed soon.
# Please remove all calls to make_sklearn_compat from your code
# as they are no longer needed


def make_sklearn_compat(op):
    """This is a deprecated method for backward compatibility and will be removed soon"""
    warnings.warn(
        "sklearn_compat.make_sklearn_compat exists for backwards compatibility and will be removed soon",
        DeprecationWarning,
    )
    return op


def sklearn_compat_clone(impl: Any) -> Any:
    """This is a deprecated method for backward compatibility and will be removed soon.
    call lale.operators.clone (or scikit-learn clone) instead"""
    warnings.warn(
        "sklearn_compat.sklearn_compat_clone exists for backwards compatibility and will be removed soon",
        DeprecationWarning,
    )
    if impl is None:
        return None

    from sklearn.base import clone

    cp = clone(impl, safe=False)
    return cp



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from test import EnableSchemaValidation

import jsonschema

import lale.lib.lale
from lale.lib.lale import ConcatFeatures, IdentityWrapper, NoOp
from lale.lib.sklearn import NMF, LogisticRegression, TfidfVectorizer
from lale.settings import (
    disable_data_schema_validation,
    disable_hyperparams_schema_validation,
    set_disable_data_schema_validation,
    set_disable_hyperparams_schema_validation,
)


class TestDatasetSchemas(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from sklearn.datasets import load_iris

        with EnableSchemaValidation():
            irisArr = load_iris()
            cls._irisArr = {"X": irisArr.data, "y": irisArr.target}
            from lale.datasets import sklearn_to_pandas

            (train_X, train_y), (test_X, test_y) = sklearn_to_pandas.load_iris_df()
            cls._irisDf = {"X": train_X, "y": train_y}
            (train_X, train_y), (test_X, test_y) = sklearn_to_pandas.digits_df()
            cls._digits = {"X": train_X, "y": train_y}
            (
                (train_X, train_y),
                (test_X, test_y),
            ) = sklearn_to_pandas.california_housing_df()
            cls._housing = {"X": train_X, "y": train_y}
            from lale.datasets import openml

            (train_X, train_y), (test_X, test_y) = openml.fetch(
                "credit-g", "classification", preprocess=False
            )
            cls._creditG = {"X": train_X, "y": train_y}
            from lale.datasets import load_movie_review

            train_X, train_y = load_movie_review()
            cls._movies = {"X": train_X, "y": train_y}
            from lale.datasets.uci.uci_datasets import fetch_drugscom

            train_X, train_y, test_X, test_y = fetch_drugscom()
            cls._drugRev = {"X": train_X, "y": train_y}

    @classmethod
    def tearDownClass(cls):
        cls._irisArr = None
        cls._irisDf = None
        cls._digits = None
        cls._housing = None
        cls._creditG = None
        cls._movies = None
        cls._drugRev = None

    def test_datasets_with_own_schemas(self):
        from lale.datasets.data_schemas import to_schema
        from lale.type_checking import validate_schema_directly

        for name in [
            "irisArr",
            "irisDf",
            "digits",
            "housing",
            "creditG",
            "movies",
            "drugRev",
        ]:
            dataset = getattr(self, f"_{name}")
            data_X, data_y = dataset["X"], dataset["y"]
            schema_X, schema_y = to_schema(data_X), to_schema(data_y)
            validate_schema_directly(data_X, schema_X, subsample_array=False)
            validate_schema_directly(data_y, schema_y, subsample_array=False)

    def test_ndarray_to_schema(self):
        from lale.datasets.data_schemas import to_schema
        from lale.type_checking import validate_schema_directly

        irisArr = self._irisArr
        assert irisArr is not None
        all_X, all_y = irisArr["X"], irisArr["y"]
        assert not hasattr(all_X, "json_schema")
        all_X_schema = to_schema(all_X)
        validate_schema_directly(all_X, all_X_schema, subsample_array=False)
        assert not hasattr(all_y, "json_schema")
        all_y_schema = to_schema(all_y)
        validate_schema_directly(all_y, all_y_schema, subsample_array=False)
        all_X_expected = {
            "type": "array",
            "minItems": 150,
            "maxItems": 150,
            "items": {
                "type": "array",
                "minItems": 4,
                "maxItems": 4,
                "items": {"type": "number"},
            },
        }
        all_y_expected = {
            "type": "array",
            "minItems": 150,
            "maxItems": 150,
            "items": {"type": "integer"},
        }
        self.maxDiff = None
        self.assertEqual(all_X_schema, all_X_expected)
        self.assertEqual(all_y_schema, all_y_expected)

    def test_pandas_to_schema(self):
        import pandas as pd

        from lale.datasets.data_schemas import to_schema
        from lale.type_checking import validate_schema_directly

        irisDf = self._irisDf
        assert irisDf is not None

        train_X, train_y = irisDf["X"], irisDf["y"]
        assert isinstance(train_X, pd.DataFrame)
        assert not hasattr(train_X, "json_schema")
        train_X_schema = to_schema(train_X)
        validate_schema_directly(train_X, train_X_schema, subsample_array=False)
        assert isinstance(train_y, pd.Series)
        assert not hasattr(train_y, "json_schema")
        train_y_schema = to_schema(train_y)
        validate_schema_directly(train_y, train_y_schema, subsample_array=False)
        train_X_expected = {
            "type": "array",
            "minItems": 120,
            "maxItems": 120,
            "items": {
                "type": "array",
                "minItems": 4,
                "maxItems": 4,
                "items": [
                    {"description": "sepal length (cm)", "type": "number"},
                    {"description": "sepal width (cm)", "type": "number"},
                    {"description": "petal length (cm)", "type": "number"},
                    {"description": "petal width (cm)", "type": "number"},
                ],
            },
        }
        train_y_expected = {
            "type": "array",
            "minItems": 120,
            "maxItems": 120,
            "items": {"description": "target", "type": "integer"},
        }
        self.maxDiff = None
        self.assertEqual(train_X_schema, train_X_expected)
        self.assertEqual(train_y_schema, train_y_expected)

    def test_arff_to_schema(self):
        from lale.datasets.data_schemas import to_schema
        from lale.type_checking import validate_schema_directly

        creditG = self._creditG
        assert creditG is not None
        train_X, train_y = creditG["X"], creditG["y"]
        assert hasattr(train_X, "json_schema")
        train_X_schema = to_schema(train_X)
        validate_schema_directly(train_X, train_X_schema, subsample_array=False)
        assert hasattr(train_y, "json_schema")
        train_y_schema = to_schema(train_y)
        validate_schema_directly(train_y, train_y_schema, subsample_array=False)
        train_X_expected = {
            "type": "array",
            "minItems": 670,
            "maxItems": 670,
            "items": {
                "type": "array",
                "minItems": 20,
                "maxItems": 20,
                "items": [
                    {
                        "description": "checking_status",
                        "enum": ["<0", "0<=X<200", ">=200", "no checking"],
                    },
                    {"description": "duration", "type": "number"},
                    {
                        "description": "credit_history",
                        "enum": [
                            "no credits/all paid",
                            "all paid",
                            "existing paid",
                            "delayed previously",
                            "critical/other existing credit",
                        ],
                    },
                    {
                        "description": "purpose",
                        "enum": [
                            "new car",
                            "used car",
                            "furniture/equipment",
                            "radio/tv",
                            "domestic appliance",
                            "repairs",
                            "education",
                            "vacation",
                            "retraining",
                            "business",
                            "other",
                        ],
                    },
                    {"description": "credit_amount", "type": "number"},
                    {
                        "description": "savings_status",
                        "enum": [
                            "<100",
                            "100<=X<500",
                            "500<=X<1000",
                            ">=1000",
                            "no known savings",
                        ],
                    },
                    {
                        "description": "employment",
                        "enum": ["unemployed", "<1", "1<=X<4", "4<=X<7", ">=7"],
                    },
                    {"description": "installment_commitment", "type": "number"},
                    {
                        "description": "personal_status",
                        "enum": [
                            "male div/sep",
                            "female div/dep/mar",
                            "male single",
                            "male mar/wid",
                            "female single",
                        ],
                    },
                    {
                        "description": "other_parties",
                        "enum": ["none", "co applicant", "guarantor"],
                    },
                    {"description": "residence_since", "type": "number"},
                    {
                        "description": "property_magnitude",
                        "enum": [
                            "real estate",
                            "life insurance",
                            "car",
                            "no known property",
                        ],
                    },
                    {"description": "age", "type": "number"},
                    {
                        "description": "other_payment_plans",
                        "enum": ["bank", "stores", "none"],
                    },
                    {"description": "housing", "enum": ["rent", "own", "for free"]},
                    {"description": "existing_credits", "type": "number"},
                    {
                        "description": "job",
                        "enum": [
                            "unemp/unskilled non res",
                            "unskilled resident",
                            "skilled",
                            "high qualif/self emp/mgmt",
                        ],
                    },
                    {"description": "num_dependents", "type": "number"},
                    {"description": "own_telephone", "enum": ["none", "yes"]},
                    {"description": "foreign_worker", "enum": ["yes", "no"]},
                ],
            },
        }
        train_y_expected = {
            "type": "array",
            "minItems": 670,
            "maxItems": 670,
            "items": {"description": "class", "enum": ["good", "bad"]},
        }
        self.maxDiff = None
        self.assertEqual(train_X_schema, train_X_expected)
        self.assertEqual(train_y_schema, train_y_expected)

    def test_keep_numbers(self):
        from lale.datasets.data_schemas import to_schema
        from lale.lib.lale import Project

        creditG = self._creditG
        assert creditG is not None
        train_X = creditG["X"]
        trainable = Project(columns={"type": "number"})
        trained = trainable.fit(train_X)
        transformed = trained.transform(train_X)
        transformed_schema = to_schema(transformed)
        transformed_expected = {
            "type": "array",
            "minItems": 670,
            "maxItems": 670,
            "items": {
                "type": "array",
                "minItems": 7,
                "maxItems": 7,
                "items": [
                    {"description": "duration", "type": "number"},
                    {"description": "credit_amount", "type": "number"},
                    {"description": "installment_commitment", "type": "number"},
                    {"description": "residence_since", "type": "number"},
                    {"description": "age", "type": "number"},
                    {"description": "existing_credits", "type": "number"},
                    {"description": "num_dependents", "type": "number"},
                ],
            },
        }
        self.maxDiff = None
        self.assertEqual(transformed_schema, transformed_expected)

    def test_keep_non_numbers(self):
        with EnableSchemaValidation():
            from lale.datasets.data_schemas import to_schema
            from lale.lib.lale import Project

            creditG = self._creditG
            assert creditG is not None
            train_X = creditG["X"]
            trainable = Project(columns={"not": {"type": "number"}})
            trained = trainable.fit(train_X)
            transformed = trained.transform(train_X)
            transformed_schema = to_schema(transformed)
            transformed_expected = {
                "type": "array",
                "minItems": 670,
                "maxItems": 670,
                "items": {
                    "type": "array",
                    "minItems": 13,
                    "maxItems": 13,
                    "items": [
                        {
                            "description": "checking_status",
                            "enum": ["<0", "0<=X<200", ">=200", "no checking"],
                        },
                        {
                            "description": "credit_history",
                            "enum": [
                                "no credits/all paid",
                                "all paid",
                                "existing paid",
                                "delayed previously",
                                "critical/other existing credit",
                            ],
                        },
                        {
                            "description": "purpose",
                            "enum": [
                                "new car",
                                "used car",
                                "furniture/equipment",
                                "radio/tv",
                                "domestic appliance",
                                "repairs",
                                "education",
                                "vacation",
                                "retraining",
                                "business",
                                "other",
                            ],
                        },
                        {
                            "description": "savings_status",
                            "enum": [
                                "<100",
                                "100<=X<500",
                                "500<=X<1000",
                                ">=1000",
                                "no known savings",
                            ],
                        },
                        {
                            "description": "employment",
                            "enum": ["unemployed", "<1", "1<=X<4", "4<=X<7", ">=7"],
                        },
                        {
                            "description": "personal_status",
                            "enum": [
                                "male div/sep",
                                "female div/dep/mar",
                                "male single",
                                "male mar/wid",
                                "female single",
                            ],
                        },
                        {
                            "description": "other_parties",
                            "enum": ["none", "co applicant", "guarantor"],
                        },
                        {
                            "description": "property_magnitude",
                            "enum": [
                                "real estate",
                                "life insurance",
                                "car",
                                "no known property",
                            ],
                        },
                        {
                            "description": "other_payment_plans",
                            "enum": ["bank", "stores", "none"],
                        },
                        {"description": "housing", "enum": ["rent", "own", "for free"]},
                        {
                            "description": "job",
                            "enum": [
                                "unemp/unskilled non res",
                                "unskilled resident",
                                "skilled",
                                "high qualif/self emp/mgmt",
                            ],
                        },
                        {"description": "own_telephone", "enum": ["none", "yes"]},
                        {"description": "foreign_worker", "enum": ["yes", "no"]},
                    ],
                },
            }
            self.maxDiff = None
            self.assertEqual(transformed_schema, transformed_expected)

    def test_input_schema_fit(self):
        self.maxDiff = None
        self.assertEqual(
            LogisticRegression.input_schema_fit(),
            LogisticRegression.get_schema("input_fit"),
        )
        self.assertEqual(
            (NMF >> LogisticRegression).input_schema_fit(), NMF.get_schema("input_fit")
        )
        self.assertEqual(
            IdentityWrapper(op=LogisticRegression).input_schema_fit(),
            LogisticRegression.get_schema("input_fit"),
        )
        actual = (TfidfVectorizer | NMF).input_schema_fit()
        expected = {
            "anyOf": [
                {
                    "type": "object",
                    "required": ["X"],
                    "additionalProperties": False,
                    "properties": {
                        "X": {
                            "anyOf": [
                                {"type": "array", "items": {"type": "string"}},
                                {
                                    "type": "array",
                                    "items": {
                                        "type": "array",
                                        "minItems": 1,
                                        "maxItems": 1,
                                        "items": {"type": "string"},
                                    },
                                },
                            ]
                        },
                        "y": {},
                    },
                },
                {
                    "type": "object",
                    "required": ["X"],
                    "additionalProperties": False,
                    "properties": {
                        "X": {
                            "type": "array",
                            "items": {
                                "type": "array",
                                "items": {"type": "number", "minimum": 0.0},
                            },
                        },
                        "y": {},
                    },
                },
            ]
        }
        self.assertEqual(actual, expected)

    def test_transform_schema_NoOp(self):
        with EnableSchemaValidation():
            from lale.datasets.data_schemas import to_schema

            for ds in [
                self._irisArr,
                self._irisDf,
                self._digits,
                self._housing,
                self._creditG,
                self._movies,
                self._drugRev,
            ]:
                assert ds is not None
                s_input = to_schema(ds["X"])
                s_output = NoOp.transform_schema(s_input)
                self.assertIs(s_input, s_output)

    def test_transform_schema_pipeline(self):
        with EnableSchemaValidation():
            from lale.datasets.data_schemas import to_schema

            pipeline = NMF >> LogisticRegression
            digits = self._digits
            assert digits is not None
            input_schema = to_schema(digits["X"])
            transformed_schema = pipeline.transform_schema(input_schema)
            transformed_expected = {
                "description": "Probability of the sample for each class in the model.",
                "type": "array",
                "items": {"type": "array", "items": {"type": "number"}},
            }
            self.maxDiff = None
            self.assertEqual(transformed_schema, transformed_expected)

    def test_transform_schema_choice(self):
        with EnableSchemaValidation():
            from lale.datasets.data_schemas import to_schema

            choice = NMF | LogisticRegression
            digits = self._digits
            assert digits is not None
            input_schema = to_schema(digits["X"])
            transformed_schema = choice.transform_schema(input_schema)
            transformed_expected = {
                "type": "array",
                "items": {"type": "array", "items": {"type": "number"}},
            }
            self.maxDiff = None
            self.assertEqual(transformed_schema, transformed_expected)

    def test_transform_schema_higher_order(self):
        with EnableSchemaValidation():
            from lale.datasets.data_schemas import to_schema

            inner = LogisticRegression
            outer = IdentityWrapper(op=LogisticRegression)
            digits = self._digits
            assert digits is not None
            input_schema = to_schema(digits["X"])
            transformed_inner = inner.transform_schema(input_schema)
            transformed_outer = outer.transform_schema(input_schema)
            self.maxDiff = None
            self.assertEqual(transformed_inner, transformed_outer)

    def test_transform_schema_Concat_irisArr(self):
        with EnableSchemaValidation():
            from lale.datasets.data_schemas import to_schema

            irisArr = self._irisArr
            assert irisArr is not None
            data_X, data_y = irisArr["X"], irisArr["y"]
            s_in_X, s_in_y = to_schema(data_X), to_schema(data_y)

            def check(s_actual, n_expected, s_expected):
                assert s_actual["items"]["minItems"] == n_expected, str(s_actual)
                assert s_actual["items"]["maxItems"] == n_expected, str(s_actual)
                assert s_actual["items"]["items"] == s_expected, str(s_actual)

            s_out_X = ConcatFeatures.transform_schema({"items": [s_in_X]})
            check(s_out_X, 4, {"type": "number"})
            s_out_y = ConcatFeatures.transform_schema({"items": [s_in_y]})
            check(s_out_y, 1, {"type": "integer"})
            s_out_XX = ConcatFeatures.transform_schema({"items": [s_in_X, s_in_X]})
            check(s_out_XX, 8, {"type": "number"})
            s_out_yy = ConcatFeatures.transform_schema({"items": [s_in_y, s_in_y]})
            check(s_out_yy, 2, {"type": "integer"})
            s_out_Xy = ConcatFeatures.transform_schema({"items": [s_in_X, s_in_y]})
            check(s_out_Xy, 5, {"type": "number"})
            s_out_XXX = ConcatFeatures.transform_schema(
                {"items": [s_in_X, s_in_X, s_in_X]}
            )
            check(s_out_XXX, 12, {"type": "number"})

    def test_transform_schema_Concat_irisDf(self):
        with EnableSchemaValidation():
            from lale.datasets.data_schemas import to_schema

            irisDf = self._irisDf
            assert irisDf is not None
            data_X, data_y = irisDf["X"], irisDf["y"]
            s_in_X, s_in_y = to_schema(data_X), to_schema(data_y)

            def check(s_actual, n_expected, s_expected):
                assert s_actual["items"]["minItems"] == n_expected, str(s_actual)
                assert s_actual["items"]["maxItems"] == n_expected, str(s_actual)
                assert s_actual["items"]["items"] == s_expected, str(s_actual)

            s_out_X = ConcatFeatures.transform_schema({"items": [s_in_X]})
            check(s_out_X, 4, {"type": "number"})
            s_out_y = ConcatFeatures.transform_schema({"items": [s_in_y]})
            check(s_out_y, 1, {"description": "target", "type": "integer"})
            s_out_XX = ConcatFeatures.transform_schema({"items": [s_in_X, s_in_X]})
            check(s_out_XX, 8, {"type": "number"})
            s_out_yy = ConcatFeatures.transform_schema({"items": [s_in_y, s_in_y]})
            check(s_out_yy, 2, {"type": "integer"})
            s_out_Xy = ConcatFeatures.transform_schema({"items": [s_in_X, s_in_y]})
            check(s_out_Xy, 5, {"type": "number"})
            s_out_XXX = ConcatFeatures.transform_schema(
                {"items": [s_in_X, s_in_X, s_in_X]}
            )
            check(s_out_XXX, 12, {"type": "number"})

    def test_lr_with_all_datasets(self):
        with EnableSchemaValidation():
            should_succeed = ["irisArr", "irisDf", "digits", "housing"]
            should_fail = ["creditG", "movies", "drugRev"]
            for name in should_succeed:
                dataset = getattr(self, f"_{name}")
                LogisticRegression.validate_schema(**dataset)
            for name in should_fail:
                dataset = getattr(self, f"_{name}")
                with self.assertRaises(ValueError):
                    LogisticRegression.validate_schema(**dataset)

    def test_project_with_all_datasets(self):
        with EnableSchemaValidation():
            should_succeed = [
                "irisArr",
                "irisDf",
                "digits",
                "housing",
                "creditG",
                "drugRev",
            ]
            should_fail = ["movies"]
            for name in should_succeed:
                dataset = getattr(self, f"_{name}")
                lale.lib.lale.Project.validate_schema(**dataset)
            for name in should_fail:
                dataset = getattr(self, f"_{name}")
                with self.assertRaises(ValueError):
                    lale.lib.lale.Project.validate_schema(**dataset)

    def test_nmf_with_all_datasets(self):
        with EnableSchemaValidation():
            should_succeed = ["digits"]
            should_fail = [
                "irisArr",
                "irisDf",
                "housing",
                "creditG",
                "movies",
                "drugRev",
            ]
            for name in should_succeed:
                dataset = getattr(self, f"_{name}")
                NMF.validate_schema(**dataset)
            for name in should_fail:
                dataset = getattr(self, f"_{name}")
                with self.assertRaises(ValueError):
                    NMF.validate_schema(**dataset)

    def test_tfidf_with_all_datasets(self):
        with EnableSchemaValidation():
            should_succeed = ["movies"]
            should_fail = [
                "irisArr",
                "irisDf",
                "digits",
                "housing",
                "creditG",
                "drugRev",
            ]
            for name in should_succeed:
                dataset = getattr(self, f"_{name}")
                TfidfVectorizer.validate_schema(**dataset)
            for name in should_fail:
                dataset = getattr(self, f"_{name}")
                with self.assertRaises(ValueError):
                    TfidfVectorizer.validate_schema(**dataset)

    def test_decision_function_binary(self):
        from lale.lib.lale import Project

        creditG = self._creditG
        assert creditG is not None
        train_X, train_y = creditG["X"], creditG["y"]
        trainable = Project(columns={"type": "number"}) >> LogisticRegression()
        trained = trainable.fit(train_X, train_y)
        _ = trained.decision_function(train_X)


class TestErrorMessages(unittest.TestCase):
    def test_wrong_cont(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError) as cm:
                LogisticRegression(C=-1)
            summary = cm.exception.message.split("\n")[0]
            self.assertEqual(
                summary,
                "Invalid configuration for LogisticRegression(C=-1) due to invalid value C=-1.",
            )
            fix1 = cm.exception.message.split("\n")[2]
            self.assertRegex(fix1, "C=1.0")

    def test_fixes2(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError) as cm:
                LogisticRegression(
                    penalty="l1",
                    solver="liblinear",
                    multi_class="multinomial",
                    dual=True,
                )
            summary = cm.exception.message.split("\n")[0]
            self.assertRegex(
                summary,
                "Invalid configuration for LogisticRegression(.*) due to constraint",
            )
            fix1 = cm.exception.message.split("\n")[2]
            fix2 = cm.exception.message.split("\n")[3]
            # we don't care what order they are in
            self.assertRegex(fix1 + fix2, "penalty='l2', multi_class='auto'")
            self.assertRegex(fix1 + fix2, "multi_class='auto', dual=False")

    def test_wrong_cat(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError) as cm:
                LogisticRegression(solver="adam")
            summary = cm.exception.message.split("\n")[0]
            self.assertEqual(
                summary,
                "Invalid configuration for LogisticRegression(solver='adam') due to invalid value solver=adam.",
            )

    def test_unknown_arg(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError) as cm:
                LogisticRegression(activation="relu")
            summary = cm.exception.message.split("\n")[0]
            self.assertEqual(
                summary,
                "Invalid configuration for LogisticRegression(activation='relu') due to argument 'activation' was unexpected.",
            )
            fix1 = cm.exception.message.split("\n")[1]
            self.assertRegex(fix1, "remove unknown key 'activation'")

    def test_constraint(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError) as cm:
                LogisticRegression(solver="sag", penalty="l1")
            summary = cm.exception.message.split("\n")[0]
            self.assertEqual(
                summary,
                "Invalid configuration for LogisticRegression(solver='sag', penalty='l1') due to constraint the newton-cg, sag, and lbfgs solvers support only l2 or no penalties.",
            )
            fix1 = cm.exception.message.split("\n")[2]
            self.assertRegex(fix1, "set penalty='l2'")

    def test_unknown_arg_and_constraint2(self):
        with EnableSchemaValidation():
            with EnableSchemaValidation():
                with self.assertRaises(jsonschema.ValidationError) as cm:
                    LogisticRegression(
                        activation="relu",
                        penalty="l1",
                        solver="liblinear",
                        multi_class="multinomial",
                        dual=True,
                    )
                summary = cm.exception.message.split("\n")[0]

                self.assertRegex(
                    summary,
                    "Invalid configuration for LogisticRegression.*due to argument 'activation' was unexpected.",
                )

                fix1 = cm.exception.message.split("\n")[2]
                fix2 = cm.exception.message.split("\n")[3]
                # we don't care what order they are in
                self.assertRegex(
                    fix1 + fix2, "remove unknown key 'activation'.*set.*penalty='l2'"
                )
                self.assertRegex(
                    fix1 + fix2, "remove unknown key 'activation'.*set.*dual=False"
                )
                self.assertRegex(
                    fix1 + fix2,
                    "remove unknown key 'activation'.*set.*multi_class='auto'",
                )

    def test_unknown_arg_and_constraint(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError) as cm:
                LogisticRegression(activation="relu", solver="sag", penalty="l1")
            summary = cm.exception.message.split("\n")[0]
            self.assertRegex(
                summary,
                "Invalid configuration for LogisticRegression.*due to argument 'activation' was unexpected.",
            )
            fix1 = cm.exception.message.split("\n")[2]
            self.assertRegex(fix1, "remove unknown key 'activation'.*set penalty='l2'")


class TestHyperparamConstraints(unittest.TestCase):
    def setUp(self):
        import scipy.sparse
        import sklearn.datasets

        data = sklearn.datasets.load_iris()
        X, y = data.data, data.target

        sparse_X = scipy.sparse.csr_matrix(X)
        self.sparse_X = sparse_X
        self.X = X
        self.y = y

        self.regression_X, self.regression_y = sklearn.datasets.make_regression(
            n_features=4, n_informative=2, random_state=0, shuffle=False
        )

    def test_bagging_classifier(self):
        import sklearn

        from lale.lib.sklearn import BaggingClassifier

        bad_hyperparams = {"bootstrap": False, "oob_score": True}
        trainable = sklearn.ensemble.BaggingClassifier(**bad_hyperparams)

        with self.assertRaisesRegex(
            ValueError, "Out of bag estimation only available if bootstrap=True"
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                BaggingClassifier(**bad_hyperparams)

    def test_bagging_classifier_2(self):
        import sklearn

        from lale.lib.sklearn import BaggingClassifier

        bad_hyperparams = {"warm_start": True, "oob_score": True}
        trainable = sklearn.ensemble.BaggingClassifier(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Out of bag estimate only available if warm_start=False"
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                BaggingClassifier(**bad_hyperparams)

    def test_bagging_regressor(self):
        import sklearn

        from lale.lib.sklearn import BaggingRegressor

        bad_hyperparams = {"bootstrap": False, "oob_score": True}
        trainable = sklearn.ensemble.BaggingRegressor(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Out of bag estimation only available if bootstrap=True"
        ):
            trainable.fit(self.regression_X, self.regression_y)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                BaggingRegressor(**bad_hyperparams)

    def test_bagging_regressor_2(self):
        import sklearn

        from lale.lib.sklearn import BaggingRegressor

        bad_hyperparams = {"warm_start": True, "oob_score": True}
        trainable = sklearn.ensemble.BaggingRegressor(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Out of bag estimate only available if warm_start=False"
        ):
            trainable.fit(self.regression_X, self.regression_y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                BaggingRegressor(**bad_hyperparams)

    def test_extra_trees_classifier(self):
        import sklearn

        from lale.lib.sklearn import ExtraTreesClassifier

        bad_hyperparams = {"bootstrap": False, "oob_score": True}
        trainable = sklearn.ensemble.ExtraTreesClassifier(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Out of bag estimation only available if bootstrap=True"
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                ExtraTreesClassifier(**bad_hyperparams)

    def test_extra_trees_regressor(self):
        import sklearn

        from lale.lib.sklearn import ExtraTreesRegressor

        bad_hyperparams = {"bootstrap": False, "oob_score": True}
        trainable = sklearn.ensemble.ExtraTreesRegressor(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Out of bag estimation only available if bootstrap=True"
        ):
            trainable.fit(self.regression_X, self.regression_y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                ExtraTreesRegressor(**bad_hyperparams)

    def test_function_transformer(self):
        import sklearn

        from lale.lib.sklearn import FunctionTransformer

        bad_hyperparams = {"validate": True, "accept_sparse": False}
        bad_X = self.sparse_X
        y = self.y

        trainable = sklearn.preprocessing.FunctionTransformer(**bad_hyperparams)
        with self.assertRaisesRegex(
            TypeError, "A sparse matrix was passed, but dense data is required."
        ):
            trainable.fit(bad_X, self.y)

        trainable = FunctionTransformer(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)

    def test_linear_svc_1(self):
        import sklearn

        from lale.lib.sklearn import LinearSVC

        bad_hyperparams = {"penalty": "l1", "loss": "hinge", "multi_class": "ovr"}
        trainable = sklearn.svm.LinearSVC(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError,
            "The combination of penalty='l1' and loss='hinge' is not supported",
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                LinearSVC(**bad_hyperparams)

    def test_linear_svc_2(self):
        import sklearn

        from lale.lib.sklearn import LinearSVC

        bad_hyperparams = {
            "penalty": "l2",
            "loss": "hinge",
            "dual": False,
            "multi_class": "ovr",
        }
        trainable = sklearn.svm.LinearSVC(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError,
            "The combination of penalty='l2' and loss='hinge' are not supported when dual=False",
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                LinearSVC(**bad_hyperparams)

    def test_linear_svc_3(self):
        import sklearn

        from lale.lib.sklearn import LinearSVC

        bad_hyperparams = {
            "penalty": "l1",
            "loss": "squared_hinge",
            "dual": True,
            "multi_class": "ovr",
        }
        trainable = sklearn.svm.LinearSVC(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError,
            "The combination of penalty='l1' and loss='squared_hinge' are not supported when dual=True",
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                LinearSVC(**bad_hyperparams)

    def test_linear_svr(self):
        import sklearn

        from lale.lib.sklearn import LinearSVR

        bad_hyperparams = {"loss": "epsilon_insensitive", "dual": False}
        trainable = sklearn.svm.LinearSVR(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError,
            "The combination of penalty='l2' and loss='epsilon_insensitive' are not supported when dual=False",
        ):
            trainable.fit(self.regression_X, self.regression_y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                LinearSVR(**bad_hyperparams)

    def test_logistic_regression_1(self):
        import sklearn

        from lale.lib.sklearn import LogisticRegression

        bad_hyperparams = {"solver": "liblinear", "penalty": "none"}
        trainable = sklearn.linear_model.LogisticRegression(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "penalty='none' is not supported for the liblinear solver"
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                LogisticRegression(**bad_hyperparams)

    def test_logistic_regression_2(self):
        import sklearn

        from lale.lib.sklearn import LogisticRegression

        bad_hyperparams = {
            "penalty": "elasticnet",
            "l1_ratio": None,
            "solver": "saga",
        }
        trainable = sklearn.linear_model.LogisticRegression(**bad_hyperparams)
        with self.assertRaisesRegex(ValueError, "l1_ratio must be between 0 and 1"):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                LogisticRegression(**bad_hyperparams)

    def test_logistic_regression_3(self):
        import sklearn

        from lale.lib.sklearn import LogisticRegression

        bad_hyperparams = {
            "penalty": "elasticnet",
            "solver": "liblinear",
            "l1_ratio": 0.5,
        }
        trainable = sklearn.linear_model.LogisticRegression(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Only 'saga' solver supports elasticnet penalty"
        ):
            trainable.fit(self.X, self.y)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                LogisticRegression(**bad_hyperparams)

    def test_missing_indicator(self):
        import sklearn

        from lale.lib.sklearn import MissingIndicator

        bad_X = self.sparse_X
        y = self.y

        bad_hyperparams = {"missing_values": 0}

        trainable = sklearn.impute.MissingIndicator(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Sparse input with missing_values=0 is not supported."
        ):
            trainable.fit(bad_X, self.y)

        trainable = MissingIndicator(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)

    def test_one_hot_encoder(self):
        import sklearn

        from lale.lib.sklearn import OneHotEncoder

        bad_hyperparams = {"drop": "first", "handle_unknown": "ignore"}
        trainable = sklearn.preprocessing.OneHotEncoder(**bad_hyperparams)
        if sklearn.__version__ < "1.0":
            with self.assertRaisesRegex(
                ValueError,
                "`handle_unknown` must be 'error' when the drop parameter is specified",
            ):
                trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            if sklearn.__version__ < "1.0":
                with self.assertRaises(jsonschema.ValidationError):
                    OneHotEncoder(**bad_hyperparams)
            else:
                OneHotEncoder(**bad_hyperparams)

    def test_ordinal_encoder_1(self):
        import sklearn

        from lale.lib.sklearn import OrdinalEncoder

        if sklearn.__version__ >= "0.24.1":
            bad_hyperparams = {
                "handle_unknown": "use_encoded_value",
                "unknown_value": None,
            }
            trainable = sklearn.preprocessing.OrdinalEncoder(**bad_hyperparams)
            with self.assertRaisesRegex(
                TypeError,
                "unknown_value should be an integer or np.nan when handle_unknown is 'use_encoded_value'",
            ):
                trainable.fit(self.X, self.y)

            with EnableSchemaValidation():
                with self.assertRaises(jsonschema.ValidationError):
                    OrdinalEncoder(**bad_hyperparams)

    def test_ordinal_encoder_2(self):
        import sklearn

        from lale.lib.sklearn import OrdinalEncoder

        if sklearn.__version__ >= "0.24.1":
            bad_hyperparams = {"handle_unknown": "error", "unknown_value": 1}
            trainable = sklearn.preprocessing.OrdinalEncoder(**bad_hyperparams)
            with self.assertRaisesRegex(
                TypeError,
                "unknown_value should only be set when handle_unknown is 'use_encoded_value'",
            ):
                trainable.fit(self.X, self.y)

            with EnableSchemaValidation():
                with self.assertRaises(jsonschema.ValidationError):
                    OrdinalEncoder(**bad_hyperparams)

    def test_random_forest_classifier(self):
        import sklearn

        from lale.lib.sklearn import RandomForestClassifier

        bad_hyperparams = {"bootstrap": False, "oob_score": True}
        trainable = sklearn.ensemble.RandomForestClassifier(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Out of bag estimation only available if bootstrap=True"
        ):
            trainable.fit(self.X, self.y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                RandomForestClassifier(**bad_hyperparams)

    def test_random_forest_regressor(self):
        import sklearn

        from lale.lib.sklearn import RandomForestRegressor

        bad_hyperparams = {"bootstrap": False, "oob_score": True}
        trainable = sklearn.ensemble.RandomForestRegressor(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "Out of bag estimation only available if bootstrap=True"
        ):
            trainable.fit(self.regression_X, self.regression_y)

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                RandomForestRegressor(**bad_hyperparams)

    def test_ridge_1(self):
        import sklearn

        from lale.lib.sklearn import Ridge

        bad_X = self.sparse_X
        y = self.y

        bad_hyperparams = {"fit_intercept": True, "solver": "lsqr"}
        trainable = sklearn.linear_model.Ridge(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError, "does not support fitting the intercept on sparse data."
        ):
            trainable.fit(bad_X, self.y)

        trainable = Ridge(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)

    def test_ridge_2(self):
        import sklearn

        from lale.lib.sklearn import Ridge

        bad_X = self.sparse_X
        y = self.y

        bad_hyperparams = {"solver": "svd", "fit_intercept": False}
        trainable = sklearn.linear_model.Ridge(**bad_hyperparams)
        with self.assertRaisesRegex(
            TypeError, "SVD solver does not support sparse inputs currently"
        ):
            trainable.fit(bad_X, self.y)

        trainable = Ridge(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)

    def test_robust_scaler(self):
        import sklearn

        from lale.lib.sklearn import RobustScaler

        bad_X = self.sparse_X
        y = self.y

        bad_hyperparams = {"with_centering": True}
        trainable = sklearn.preprocessing.RobustScaler(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError,
            "Cannot center sparse matrices: use `with_centering=False` instead.",
        ):
            trainable.fit(bad_X, self.y)

        trainable = RobustScaler(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)

    def test_simple_imputer(self):
        import sklearn

        from lale.lib.sklearn import SimpleImputer

        bad_X = self.sparse_X
        y = self.y

        bad_hyperparams = {"missing_values": 0}
        trainable = sklearn.impute.SimpleImputer(**bad_hyperparams)
        with self.assertRaisesRegex(
            ValueError,
            "Imputation not possible when missing_values == 0 and input is sparse.",
        ):
            trainable.fit(bad_X, self.y)

        trainable = SimpleImputer(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)

    def test_svc(self):
        import sklearn

        from lale.lib.sklearn import SVC

        bad_X = self.sparse_X
        y = self.y

        bad_hyperparams = {"kernel": "precomputed"}
        trainable = sklearn.svm.SVC(**bad_hyperparams)
        with self.assertRaisesRegex(
            TypeError, "Sparse precomputed kernels are not supported."
        ):
            trainable.fit(bad_X, self.y)

        trainable = SVC(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)

    def test_svr(self):
        import sklearn

        from lale.lib.sklearn import SVR

        bad_X = self.sparse_X
        y = self.y

        bad_hyperparams = {"kernel": "precomputed"}
        trainable = sklearn.svm.SVR(**bad_hyperparams)
        with self.assertRaisesRegex(
            TypeError, "Sparse precomputed kernels are not supported."
        ):
            trainable.fit(bad_X, self.y)

        trainable = SVR(**bad_hyperparams)
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                trainable.fit(bad_X, y)


class TestSchemaValidation(unittest.TestCase):
    def test_any(self):
        from lale.type_checking import is_subschema

        num_schema = {"type": "number"}
        any_schema = {"laleType": "Any"}
        jsonschema.validate(42, num_schema)
        jsonschema.validate(42, any_schema)
        self.assertTrue(is_subschema(num_schema, any_schema))
        self.assertTrue(is_subschema(any_schema, num_schema))

    def test_bool_label(self):
        import pandas as pd

        data_records = [
            {
                "IS_TENT": False,
                "GENDER": "M",
                "AGE": 20,
                "MARITAL_STATUS": "Single",
                "PROFESSION": "Sales",
            },
            {
                "IS_TENT": False,
                "GENDER": "M",
                "AGE": 20,
                "MARITAL_STATUS": "Single",
                "PROFESSION": "Sales",
            },
            {
                "IS_TENT": False,
                "GENDER": "F",
                "AGE": 37,
                "MARITAL_STATUS": "Single",
                "PROFESSION": "Other",
            },
            {
                "IS_TENT": False,
                "GENDER": "M",
                "AGE": 42,
                "MARITAL_STATUS": "Married",
                "PROFESSION": "Other",
            },
            {
                "IS_TENT": True,
                "GENDER": "F",
                "AGE": 24,
                "MARITAL_STATUS": "Married",
                "PROFESSION": "Retail",
            },
            {
                "IS_TENT": False,
                "GENDER": "F",
                "AGE": 24,
                "MARITAL_STATUS": "Married",
                "PROFESSION": "Retail",
            },
            {
                "IS_TENT": False,
                "GENDER": "M",
                "AGE": 29,
                "MARITAL_STATUS": "Single",
                "PROFESSION": "Retail",
            },
            {
                "IS_TENT": False,
                "GENDER": "M",
                "AGE": 29,
                "MARITAL_STATUS": "Single",
                "PROFESSION": "Retail",
            },
            {
                "IS_TENT": True,
                "GENDER": "M",
                "AGE": 43,
                "MARITAL_STATUS": "Married",
                "PROFESSION": "Trades",
            },
            {
                "IS_TENT": False,
                "GENDER": "M",
                "AGE": 43,
                "MARITAL_STATUS": "Married",
                "PROFESSION": "Trades",
            },
        ]
        df = pd.DataFrame.from_records(data_records)
        X = df.drop(["IS_TENT"], axis=1).values
        y = df["IS_TENT"].values
        from lale.lib.sklearn import GradientBoostingClassifier as Clf
        from lale.lib.sklearn import OneHotEncoder as Enc

        trainable = Enc() >> Clf()
        _ = trainable.fit(X, y)


class TestWithScorer(unittest.TestCase):
    def test_bare_array(self):
        import sklearn.datasets
        import sklearn.metrics
        from numpy import ndarray

        from lale.datasets.data_schemas import NDArrayWithSchema

        X, y = sklearn.datasets.load_iris(return_X_y=True)
        self.assertIsInstance(X, ndarray)
        self.assertIsInstance(y, ndarray)
        self.assertNotIsInstance(X, NDArrayWithSchema)
        self.assertNotIsInstance(y, NDArrayWithSchema)
        trainable = LogisticRegression()
        trained = trainable.fit(X, y)
        scorer = sklearn.metrics.make_scorer(sklearn.metrics.accuracy_score)
        out = scorer(trained, X, y)
        self.assertIsInstance(out, float)
        self.assertNotIsInstance(out, NDArrayWithSchema)


class TestDisablingSchemaValidation(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)

    def test_disable_schema_validation_individual_op(self):
        existing_flag = disable_data_schema_validation
        set_disable_data_schema_validation(True)
        import lale.schemas as schemas
        from lale.lib.sklearn import PCA

        pca_input = schemas.Object(
            X=schemas.AnyOf(
                [
                    schemas.Array(schemas.Array(schemas.String())),
                    schemas.Array(schemas.String()),
                ]
            )
        )

        foo = PCA.customize_schema(input_fit=pca_input)

        pca_output = schemas.Object(
            X=schemas.AnyOf(
                [
                    schemas.Array(schemas.Array(schemas.String())),
                    schemas.Array(schemas.String()),
                ]
            )
        )

        foo = foo.customize_schema(output_transform=pca_output)

        abc = foo()
        trained_pca = abc.fit(self.X_train)
        trained_pca.transform(self.X_test)
        set_disable_data_schema_validation(existing_flag)

    def test_enable_schema_validation_individual_op(self):
        with EnableSchemaValidation():
            import lale.schemas as schemas
            from lale.lib.sklearn import PCA

            pca_input = schemas.Object(
                X=schemas.AnyOf(
                    [
                        schemas.Array(schemas.Array(schemas.String())),
                        schemas.Array(schemas.String()),
                    ]
                )
            )

            foo = PCA.customize_schema(input_fit=pca_input)

            pca_output = schemas.Object(
                X=schemas.AnyOf(
                    [
                        schemas.Array(schemas.Array(schemas.String())),
                        schemas.Array(schemas.String()),
                    ]
                )
            )

            foo = foo.customize_schema(output_transform=pca_output)

            abc = foo()
            with self.assertRaises(ValueError):
                trained_pca = abc.fit(self.X_train)
                trained_pca.transform(self.X_test)

    def test_disable_schema_validation_pipeline(self):
        existing_flag = disable_data_schema_validation
        set_disable_data_schema_validation(True)
        import lale.schemas as schemas
        from lale.lib.sklearn import PCA, LogisticRegression

        lr_input = schemas.Object(
            required=["X", "y"],
            X=schemas.AnyOf(
                [
                    schemas.Array(schemas.Array(schemas.String())),
                    schemas.Array(schemas.String()),
                ]
            ),
            y=schemas.Array(schemas.String()),
        )

        foo = LogisticRegression.customize_schema(input_fit=lr_input)
        abc = foo()
        pipeline = PCA() >> abc
        trained_pipeline = pipeline.fit(self.X_train, self.y_train)
        trained_pipeline.predict(self.X_test)
        set_disable_data_schema_validation(existing_flag)

    def test_enable_schema_validation_pipeline(self):
        with EnableSchemaValidation():
            import lale.schemas as schemas
            from lale.lib.sklearn import PCA, LogisticRegression

            lr_input = schemas.Object(
                required=["X", "y"],
                X=schemas.AnyOf(
                    [
                        schemas.Array(schemas.Array(schemas.String())),
                        schemas.Array(schemas.String()),
                    ]
                ),
                y=schemas.Array(schemas.String()),
            )

            foo = LogisticRegression.customize_schema(input_fit=lr_input)
            abc = foo()
            pipeline = PCA() >> abc
            with self.assertRaises(ValueError):
                trained_pipeline = pipeline.fit(self.X_train, self.y_train)
                trained_pipeline.predict(self.X_test)

    def test_disable_enable_hyperparam_validation(self):
        from lale.lib.sklearn import PCA

        existing_flag = disable_hyperparams_schema_validation
        set_disable_hyperparams_schema_validation(True)
        PCA(n_components=True)
        set_disable_hyperparams_schema_validation(False)
        with self.assertRaises(jsonschema.ValidationError):
            PCA(n_components=True)
        set_disable_hyperparams_schema_validation(existing_flag)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
from sklearn.ensemble import IsolationForest as SKLModel

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": """Isolation Forest Algorithm.
Return the anomaly score of each sample using the IsolationForest algorithm.
The IsolationForest ‘isolates’ observations by randomly selecting a feature and then randomly selecting a split value between the maximum and minimum values of the selected feature.
Since recursive partitioning can be represented by a tree structure,
the number of splittings required to isolate a sample is equivalent to the path length from the root node to the terminating node.
This path length, averaged over a forest of such random trees, is a measure of normality and our decision function.
Random partitioning produces noticeably shorter paths for anomalies. Hence, when a forest of random trees collectively
produce shorter path lengths for particular samples, they are highly likely to be anomalies.""",
    "allOf": [
        {
            "type": "object",
            "required": [
                "n_estimators",
                "max_samples",
                "contamination",
                "max_features",
                "bootstrap",
                "n_jobs",
                "behaviour",
                "random_state",
                "verbose",
                "warm_start",
            ],
            "relevantToOptimizer": [
                "n_estimators",
                "max_samples",
                "max_features",
                "bootstrap",
            ],
            "additionalProperties": False,
            "properties": {
                "n_estimators": {
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 100,
                    "distribution": "uniform",
                    "default": 100,
                    "description": "The number of base estimators in the ensemble.",
                },
                "max_samples": {
                    "description": "The number of samples to draw from X to train each base estimator.",
                    "anyOf": [
                        {
                            "description": "Draw max_samples samples.",
                            "type": "integer",
                            "minimum": 2,
                            "laleMaximum": "X/maxItems",  # number of rows
                            "forOptimizer": False,
                        },
                        {
                            "description": "Draw max_samples * X.shape[0] samples.",
                            "type": "number",
                            "minimum": 0.0,
                            "exclusiveMinimum": True,
                            "maximum": 1.0,
                            "minimumForOptimizer": 0.2,
                            "maximumForOptimizer": 1.0,
                        },
                        {
                            "description": "Draw max_samples=min(256, n_samples) samples.",
                            "enum": ["auto"],
                        },
                    ],
                    "default": "auto",
                },
                "contamination": {
                    "description": """The amount of contamination of the data set, i.e. the proportion of outliers in the data set.
Used when fitting to define the threshold on the scores of the samples.""",
                    "anyOf": [
                        {
                            "type": "number",
                            "minimum": 0.0,
                            "maximum": 0.5,
                        },
                        {"enum": ["auto"]},
                    ],
                    "default": "auto",
                },
                "max_features": {
                    "description": "The number of features to draw from X to train each base estimator.",
                    "anyOf": [
                        {
                            "description": "Draw max_features features.",
                            "type": "integer",
                            "minimum": 2,
                            "laleMaximum": "X/items/maxItems",  # number of columns
                            "forOptimizer": False,
                        },
                        {
                            "description": "Draw max_samples * X.shape[1] features.",
                            "type": "number",
                            "minimum": 0.0,
                            "exclusiveMinimum": True,
                            "maximum": 1.0,
                            "minimumForOptimizer": 0.01,
                            "maximumForOptimizer": 1.0,
                        },
                    ],
                    "default": 1.0,
                },
                "bootstrap": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether samples are drawn with (True) or without (False) replacement.",
                },
                "n_jobs": {
                    "description": "The number of jobs to run in parallel for both `fit` and `predict`.",
                    "anyOf": [
                        {
                            "description": "1 unless in joblib.parallel_backend context.",
                            "enum": [None],
                        },
                        {"description": "Use all processors.", "enum": [-1]},
                        {
                            "description": "Number of CPU cores.",
                            "type": "integer",
                            "minimum": 1,
                        },
                    ],
                    "default": None,
                },
                "behaviour": {
                    "description": "This parameter has no effect, is deprecated, and will be removed.",
                    "enum": ["deprecated"],
                    "default": "deprecated",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": """Controls the pseudo-randomness of the selection of the feature and split values for each branching step and each tree in the forest.
If int, random_state is the seed used by the random number generator""",
                },
                "verbose": {
                    "type": "integer",
                    "default": 0,
                    "description": "Controls the verbosity of the tree building process.",
                },
                "warm_start": {
                    "type": "boolean",
                    "default": False,
                    "description": "When set to True, reuse the solution of the previous call to fit and add more estimators to the ensemble, otherwise, just fit a whole new ensemble.",
                },
            },
        }
    ],
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The training input samples. Sparse matrices are accepted only if",
        },
        "y": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "number"},
                    "description": "The target values (class labels in classification, real numbers in",
                },
                {"enum": [None]},
            ]
        },
        "sample_weight": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "number"},
                },
                {"enum": [None]},
            ],
            "description": "Sample weights. If None, then samples are equally weighted.",
        },
    },
}

_input_predict_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
        },
    },
}

_output_predict_schema = {
    "type": "array",
    "items": {"type": "number"},
}

_input_decision_function_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        }
    },
}

_output_decision_function_schema = {
    "type": "array",
    "items": {"type": "number"},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Isolation forest`_ from scikit-learn for getting the anomaly score of each sample using the IsolationForest algorithm.

.. _`Isolation forest`: https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.isolation_forest.html",
    "import_from": "sklearn.ensemble",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_decision_function": _input_decision_function_schema,
        "output_decision_function": _output_decision_function_schema,
    },
}

IsolationForest = lale.operators.make_operator(SKLModel, _combined_schemas)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.22/modules/generated/sklearn.ensemble.IsolationForest.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.ensemble.IsolationForest.html
    IsolationForest = IsolationForest.customize_schema(
        behaviour=None, set_as_available=True
    )

lale.docstrings.set_docstrings(IsolationForest)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.preprocessing

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Standardize features by removing the mean and scaling to unit variance",
    "allOf": [
        {
            "type": "object",
            "required": ["copy", "with_mean", "with_std"],
            "relevantToOptimizer": ["with_mean", "with_std"],
            "additionalProperties": False,
            "properties": {
                "copy": {
                    "type": "boolean",
                    "default": True,
                    "description": "If False, try to avoid a copy and do inplace scaling instead.",
                },
                "with_mean": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, center the data before scaling.",
                },
                "with_std": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, scale the data to unit variance (or equivalently, unit standard deviation).",
                },
            },
        },
        {
            "description": "Setting `with_mean` to True does not work on sparse matrices, because centering them entails building a dense matrix which in common use cases is likely to be too large to fit in memory.",
            "anyOf": [
                {"type": "object", "properties": {"with_mean": {"enum": [False]}}},
                {"type": "object", "laleNot": "X/isSparse"},
            ],
        },
    ],
}

_input_fit_schema = {
    "description": "Compute the mean and std to be used for later scaling.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The data used to compute the mean and standard deviation",
        },
        "y": {"description": "Ignored"},
    },
}

_input_transform_schema = {
    "description": "Perform standardization by centering and scaling",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The data used to scale along the features axis.",
        },
        "copy": {
            "anyOf": [{"type": "boolean"}, {"enum": [None]}],
            "default": None,
            "description": "Copy the input X or not.",
        },
    },
}

_output_transform_schema = {
    "description": "Perform standardization by centering and scaling",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Standard scaler`_ transformer from scikit-learn.

.. _`Standard scaler`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.standard_scaler.html",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


StandardScaler = lale.operators.make_operator(
    sklearn.preprocessing.StandardScaler, _combined_schemas
)

lale.docstrings.set_docstrings(StandardScaler)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.kernel_approximation

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Hyperparameter schema for the Nystroem model from scikit-learn.",
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "kernel",
                "gamma",
                "coef0",
                "degree",
                "n_components",
                "random_state",
            ],
            "relevantToOptimizer": [
                "kernel",
                "gamma",
                "coef0",
                "degree",
                "n_components",
            ],
            "properties": {
                "kernel": {
                    "description": "Kernel map to be approximated.",
                    "anyOf": [
                        {
                            "description": "keys of sklearn.metrics.pairwise.KERNEL_PARAMS",
                            "enum": [
                                "additive_chi2",
                                "chi2",
                                "cosine",
                                "linear",
                                "poly",
                                "polynomial",
                                "rbf",
                                "laplacian",
                                "sigmoid",
                            ],
                        },
                        {"laleType": "callable", "forOptimizer": False},
                    ],
                    "default": "rbf",
                },
                "gamma": {
                    "description": "Gamma parameter.",
                    "anyOf": [
                        {"enum": [None]},
                        {
                            "type": "number",
                            "distribution": "loguniform",
                            "minimumForOptimizer": 3.0517578125e-05,
                            "maximumForOptimizer": 8,
                        },
                    ],
                    "default": None,
                },
                "coef0": {
                    "description": "Zero coefficient.",
                    "anyOf": [
                        {"enum": [None]},
                        {
                            "type": "number",
                            "minimum": (-1),
                            "distribution": "uniform",
                            "maximumForOptimizer": 1,
                        },
                    ],
                    "default": None,
                },
                "degree": {
                    "description": "Degree of the polynomial kernel.",
                    "anyOf": [
                        {"enum": [None]},
                        {
                            "type": "integer",
                            "minimumForOptimizer": 2,
                            "maximumForOptimizer": 5,
                        },
                    ],
                    "default": None,
                },
                "kernel_params": {
                    "description": "Additional parameters (keyword arguments) for kernel "
                    "function passed as callable object.",
                    "anyOf": [{"type": "object"}, {"enum": [None]}],
                    "default": None,
                },
                "n_components": {
                    "description": "Number of features to construct. How many data points will be used to construct the mapping.",
                    "type": "integer",
                    "default": 100,
                    "minimum": 1,
                    "distribution": "uniform",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 256,
                },
                "random_state": {
                    "description": "Seed of pseudo-random number generator.",
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
            },
        }
    ],
}

_input_fit_schema = {
    "description": "Input data schema for training the Nystroem model from scikit-learn.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        "y": {"description": "Target class labels; the array is over samples."},
    },
}

_input_transform_schema = {
    "description": "Input data schema for predictions using the Nystroem model from scikit-learn.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        }
    },
}

_output_transform_schema = {
    "description": "Output data schema for predictions (projected data) using the Nystroem model from scikit-learn.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Nystroem`_ transformer from scikit-learn.

.. _`Nystroem`: https://scikit-learn.org/stable/modules/generated/sklearn.kernel_approximation.Nystroem.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.nystroem.html",
    "import_from": "sklearn.kernel_approximation",
    "type": "object",
    "tags": {"pre": ["~categoricals"], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

Nystroem: lale.operators.PlannedIndividualOp
Nystroem = lale.operators.make_operator(
    sklearn.kernel_approximation.Nystroem, _combined_schemas
)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.kernel_approximation.Nystroem.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.kernel_approximation.Nystroem.html
    Nystroem = Nystroem.customize_schema(
        n_jobs={
            "anyOf": [
                {
                    "description": "1 unless in joblib.parallel_backend context.",
                    "enum": [None],
                },
                {"description": "Use all processors.", "enum": [-1]},
                {
                    "description": "Number of CPU cores.",
                    "type": "integer",
                    "minimum": 1,
                },
            ],
            "default": None,
            "description": "The number of jobs to use for the computation.",
        },
        set_as_available=True,
    )


lale.docstrings.set_docstrings(Nystroem)



from numpy import inf, nan
from sklearn.random_projection import SparseRandomProjection as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _SparseRandomProjectionImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def transform(self, X):
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for SparseRandomProjection    Reduce dimensionality through sparse random projection",
    "allOf": [
        {
            "type": "object",
            "required": [
                "n_components",
                "density",
                "eps",
                "dense_output",
                "random_state",
            ],
            "relevantToOptimizer": ["n_components", "eps", "dense_output", "density"],
            "additionalProperties": False,
            "properties": {
                "n_components": {
                    "XXX TODO XXX": "int or 'auto', optional (default = 'auto')",
                    "description": "Dimensionality of the target projection space",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 2,
                            "maximumForOptimizer": 256,
                            "distribution": "uniform",
                        },
                        {"enum": ["auto"]},
                    ],
                    "default": "auto",
                },
                "density": {
                    "XXX TODO XXX": "float in range ]0, 1], optional (default='auto')",
                    "description": "Ratio of non-zero component in the random projection matrix",
                    "anyOf": [
                        {"enum": ["auto"]},
                        {
                            "type": "number",
                            "minimumForOptimizer": 0,
                            "maximumForOptimizer": 1,
                            "distribution": "uniform",
                        },
                    ],
                    "default": "auto",
                },
                "eps": {
                    "XXX TODO XXX": "strictly positive float, optional, (default=0.1)",
                    "description": "Parameter to control the quality of the embedding according to the Johnson-Lindenstrauss lemma when n_components is set to 'auto'",
                    "type": "number",
                    "minimumForOptimizer": 0.001,
                    "maximumForOptimizer": 0.1,
                    "distribution": "loguniform",
                    "default": 0.1,
                },
                "dense_output": {
                    "type": "boolean",
                    "default": False,
                    "description": "If True, ensure that the output of the random projection is a dense numpy array even if the input and random projection matrix are both sparse",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Control the pseudo random number generator used to generate the matrix at fit time",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Generate a sparse random projection matrix",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "numpy array or scipy.sparse of shape [n_samples, n_features]",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Training set: only the shape is used to find optimal random matrix dimensions based on the theory referenced in the afore mentioned papers.",
        },
        "y": {"laleType": "Any", "XXX TODO XXX": "", "description": "Ignored"},
    },
}
_input_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Project the data by using matrix product with the random matrix",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "numpy array or scipy.sparse of shape [n_samples, n_features]",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "The input data to project into a smaller dimensional space.",
        }
    },
}
_output_transform_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Projected array.",
    "anyOf": [
        {
            "type": "array",
            "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
            "XXX TODO XXX": "numpy array or scipy sparse of shape [n_samples, n_components]",
        },
        {"type": "array", "items": {"type": "array", "items": {"type": "number"}}},
    ],
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.random_projection.SparseRandomProjection#sklearn-random_projection-sparserandomprojection",
    "import_from": "sklearn.random_projection",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}
SparseRandomProjection = make_operator(_SparseRandomProjectionImpl, _combined_schemas)

set_docstrings(SparseRandomProjection)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.linear_model

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "required": ["fit_intercept", "normalize", "copy_X"],
            "relevantToOptimizer": ["fit_intercept", "normalize", "copy_X"],
            "additionalProperties": False,
            "properties": {
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to calculate the intercept for this model.",
                },
                "normalize": {
                    "type": "boolean",
                    "default": False,
                    "description": "If True, the regressors X will be normalized before regression by subtracting the mean and dividing by the l2-norm.",
                },
                "copy_X": {
                    "type": "boolean",
                    "default": True,
                    "description": "If True, X will be copied; else, it may be overwritten.",
                },
                "n_jobs": {
                    "anyOf": [
                        {
                            "description": "1 unless in joblib.parallel_backend context.",
                            "enum": [None],
                        },
                        {"description": "Use all processors.", "enum": [-1]},
                        {
                            "description": "Number of CPU cores.",
                            "type": "integer",
                            "minimum": 1,
                        },
                    ],
                    "default": None,
                    "description": "The number of jobs to run in parallel.",
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        "y": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
                {"type": "array", "items": {"type": "number"}},
            ],
            "description": "Target values. Will be cast to X's dtype if necessary",
        },
        "sample_weight": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"enum": [None], "description": "Samples are equally weighted."},
            ],
            "description": "Sample weights.",
        },
    },
}

_input_predict_schema = {
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Samples.",
        }
    },
}

_output_predict_schema = {
    "description": "Returns predicted values.",
    "anyOf": [
        {"type": "array", "items": {"type": "array", "items": {"type": "number"}}},
        {"type": "array", "items": {"type": "number"}},
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Linear regression`_ linear model from scikit-learn for classification.

.. _`Linear regression`: https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.linear_regression.html",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}

LinearRegression: lale.operators.PlannedIndividualOp
LinearRegression = lale.operators.make_operator(
    sklearn.linear_model.LinearRegression, _combined_schemas
)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.LinearRegression.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.linear_model.LinearRegression.html
    LinearRegression = LinearRegression.customize_schema(
        positive={
            "type": "boolean",
            "description": "When set to True, forces the coefficients to be positive.",
            "default": False,
        },
        set_as_available=True,
    )
    LinearRegression = LinearRegression.customize_schema(
        constraint={
            "description": "Setting positive=True is only supported for dense arrays.",
            "anyOf": [
                {"type": "object", "properties": {"positive": {"enum": [False]}}},
                {"type": "object", "laleNot": "X/isSparse"},
            ],
        },
        set_as_available=True,
    )

if sklearn.__version__ >= "1.0":
    # old: https://scikit-learn.org/0.24/modules/generated/sklearn.linear_model.LinearRegression.html
    # new: https://scikit-learn.org/1.0/modules/generated/sklearn.linear_model.LinearRegression.html
    LinearRegression = LinearRegression.customize_schema(
        relevantToOptimizer=["fit_intercept", "copy_X"],
        normalize={
            "anyOf": [{"type": "boolean"}, {"enum": ["deprecated"]}],
            "description": """This parameter is ignored when fit_intercept is set to False.
If True, the regressors X will be normalized before regression by subtracting the mean and dividing by the l2-norm.
If you wish to standardize, please use StandardScaler before calling fit on an estimator with normalize=False.""",
            "default": False,
            "forOptimizer": False,
        },
        set_as_available=True,
    )

lale.docstrings.set_docstrings(LinearRegression)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

import lale.docstrings
import lale.operators

from ._common_schemas import schema_silent

try:
    import xgboost  # type: ignore

    xgboost_installed = True
except ImportError:
    xgboost_installed = False
    if TYPE_CHECKING:
        import xgboost  # type: ignore


def _rename_one_feature(name):
    mapping = {"[": "&#91;", "]": "&#93;", "<": "&lt;"}
    for old, new in mapping.items():
        name = name.replace(old, new)
    return name


def _rename_all_features(X):
    if not isinstance(X, pd.DataFrame):
        return X
    mapped = [_rename_one_feature(f) for f in X.columns]
    if list(X.columns) == mapped:
        return X
    return pd.DataFrame(data=X, columns=mapped)


class _XGBRegressorImpl:
    _wrapped_model: xgboost.XGBRegressor

    @classmethod
    def validate_hyperparams(cls, **hyperparams):
        assert xgboost_installed, """Your Python environment does not have xgboost installed. You can install it with
            pip install xgboost
        or with
            pip install 'lale[full]'"""

    def __init__(self, **hyperparams):
        self.validate_hyperparams(**hyperparams)
        self._hyperparams = hyperparams

    def fit(self, X, y, **fit_params):
        result = _XGBRegressorImpl(**self._hyperparams)
        result._wrapped_model = xgboost.XGBRegressor(**self._hyperparams)
        renamed_X = _rename_all_features(X)
        result._wrapped_model.fit(renamed_X, y, **fit_params)
        return result

    def predict(self, X, **predict_params):
        renamed_X = _rename_all_features(X)
        result = self._wrapped_model.predict(renamed_X, **predict_params)
        return result

    def predict_proba(self, X):
        return self._wrapped_model.predict_proba(X)

    def score(self, X, y):
        from sklearn.metrics import r2_score

        y_pred = self.predict(X)
        return r2_score(y, y_pred)


_hyperparams_schema = {
    "description": "Hyperparameter schema for a Lale wrapper for XGBoost.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "max_depth",
                "learning_rate",
                "n_estimators",
                "verbosity",
                "objective",
                "booster",
                "tree_method",
                "n_jobs",
                "gamma",
                "min_child_weight",
                "max_delta_step",
                "subsample",
                "colsample_bytree",
                "colsample_bylevel",
                "colsample_bynode",
                "reg_alpha",
                "reg_lambda",
                "scale_pos_weight",
                "base_score",
                "random_state",
                "missing",
            ],
            "relevantToOptimizer": [
                "max_depth",
                "learning_rate",
                "n_estimators",
                "gamma",
                "min_child_weight",
                "subsample",
                "reg_alpha",
                "reg_lambda",
            ],
            "properties": {
                "max_depth": {
                    "description": "Maximum tree depth for base learners.",
                    "type": "integer",
                    "default": 4,
                    "minimum": 0,
                    "distribution": "uniform",
                    "minimumForOptimizer": 1,
                    "maximumForOptimizer": 7,
                },
                "learning_rate": {
                    "description": """Boosting learning rate (xgb's "eta")""",
                    "type": "number",
                    "default": 0.1,
                    "distribution": "loguniform",
                    "minimumForOptimizer": 0.02,
                    "maximumForOptimizer": 1,
                },
                "n_estimators": {
                    "description": "Number of trees to fit.",
                    "type": "integer",
                    "default": 200,
                    "minimumForOptimizer": 50,
                    "maximumForOptimizer": 1000,
                },
                "verbosity": {
                    "description": "The degree of verbosity.",
                    "type": "integer",
                    "default": 1,
                    "minimum": 0,
                    "maximum": 3,
                },
                "silent": schema_silent,
                "objective": {
                    "description": "Specify the learning task and the corresponding "
                    "learning objective or a custom objective function to be used.",
                    "anyOf": [
                        {
                            "enum": [
                                "reg:linear",
                                "reg:logistic",
                                "reg:gamma",
                                "reg:tweedie",
                            ]
                        },
                        {"laleType": "callable"},
                    ],
                    "default": "reg:linear",
                },
                "booster": {
                    "description": "Specify which booster to use.",
                    "enum": ["gbtree", "gblinear", "dart"],
                    "default": "gbtree",
                },
                "tree_method": {
                    "description": """Specify which tree method to use.
Default to auto. If this parameter is set to default, XGBoost will choose the most conservative option available.
Refer to https://xgboost.readthedocs.io/en/latest/parameter.html. """,
                    "enum": ["auto", "exact", "approx", "hist", "gpu_hist"],
                    "default": "auto",
                },
                "n_jobs": {
                    "type": "integer",
                    "description": "Number of parallel threads used to run xgboost.  (replaces ``nthread``)",
                    "default": 1,
                },
                "nthread": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": None,
                    "description": "Number of parallel threads used to run xgboost.  Deprecated, please use n_jobs",
                },
                "gamma": {
                    "type": "number",
                    "description": "Minimum loss reduction required to make a further partition on a leaf node of the tree.",
                    "default": 0,
                    "minimum": 0,
                    "maximumForOptimizer": 1.0,
                },
                "min_child_weight": {
                    "type": "integer",
                    "description": "Minimum sum of instance weight(hessian) needed in a child.",
                    "default": 10,
                    "distribution": "uniform",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 20,
                },
                "max_delta_step": {
                    "type": "integer",
                    "description": "Maximum delta step we allow each tree's weight estimation to be.",
                    "default": 0,
                },
                "subsample": {
                    "type": "number",
                    "description": "Subsample ratio of the training instance.",
                    "default": 1,
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.01,
                    "maximumForOptimizer": 1.0,
                },
                "colsample_bytree": {
                    "type": "number",
                    "description": "Subsample ratio of columns when constructing each tree.",
                    "default": 1,
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "maximum": 1,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.1,
                    "maximumForOptimizer": 1.0,
                },
                "colsample_bylevel": {
                    "type": "number",
                    "description": "Subsample ratio of columns for each split, in each level.",
                    "default": 1,
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "maximum": 1,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.1,
                    "maximumForOptimizer": 1.0,
                },
                "colsample_bynode": {
                    "type": "number",
                    "description": "Subsample ratio of columns for each split.",
                    "default": 1,
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "maximum": 1,
                },
                "reg_alpha": {
                    "type": "number",
                    "description": "L1 regularization term on weights",
                    "default": 0,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0,
                    "maximumForOptimizer": 1,
                },
                "reg_lambda": {
                    "type": "number",
                    "description": "L2 regularization term on weights",
                    "default": 1,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.1,
                    "maximumForOptimizer": 1,
                },
                "scale_pos_weight": {
                    "type": "number",
                    "description": "Balancing of positive and negative weights.",
                    "default": 1,
                },
                "base_score": {
                    "type": "number",
                    "description": "The initial prediction score of all instances, global bias.",
                    "default": 0.5,
                },
                "random_state": {
                    "type": "integer",
                    "description": "Random number seed.  (replaces seed)",
                    "default": 0,
                },
                "missing": {
                    "anyOf": [
                        {
                            "type": "number",
                        },
                        {
                            "enum": [None],
                        },
                    ],
                    "default": None,
                    "description": "Value in the data which needs to be present as a missing value. If"
                    " If None, defaults to np.nan.",
                },
                "importance_type": {
                    "enum": [
                        "gain",
                        "weight",
                        "cover",
                        "total_gain",
                        "total_cover",
                        None,
                    ],
                    "default": "gain",
                    "description": "The feature importance type for the `feature_importances_` property.",
                },
                "seed": {
                    "default": None,
                    "description": "deprecated and replaced with random_state, but adding to be backward compatible. ",
                },
            },
        }
    ],
}
_input_fit_schema = {
    "description": "Fit gradient boosting classifier",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "Feature matrix",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Labels",
        },
        "sample_weight": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "number"},
                },
                {"enum": [None]},
            ],
            "description": "Weight for each instance",
            "default": None,
        },
        "eval_set": {
            "anyOf": [
                {
                    "type": "array",
                },
                {
                    "enum": [None],
                },
            ],
            "default": None,
            "description": "A list of (X, y) pairs to use as a validation set for",
        },
        "sample_weight_eval_set": {
            "anyOf": [
                {
                    "type": "array",
                },
                {
                    "enum": [None],
                },
            ],
            "default": None,
            "description": "A list of the form [L_1, L_2, ..., L_n], where each L_i is a list of",
        },
        "eval_metric": {
            "anyOf": [
                {"type": "array", "items": {"type": "string"}},
                {"type": "string"},
                {"enum": [None]},
                {"type": "object"},
            ],
            "default": None,
            "description": "If a str, should be a built-in evaluation metric to use. See",
        },
        "early_stopping_rounds": {
            "anyOf": [
                {
                    "type": "integer",
                },
                {
                    "enum": [None],
                },
            ],
            "default": None,
            "description": "Activates early stopping. Validation error needs to decrease at",
        },
        "verbose": {
            "type": "boolean",
            "description": "If `verbose` and an evaluation set is used, writes the evaluation",
            "default": True,
        },
        "xgb_model": {
            "anyOf": [{"type": "string"}, {"enum": [None]}],
            "description": "file name of stored xgb model or 'Booster' instance Xgb model to be",
            "default": None,
        },
        "callbacks": {
            "anyOf": [{"type": "array", "items": {"type": "object"}}, {"enum": [None]}],
            "default": None,
            "description": "List of callback functions that are applied at each iteration. ",
        },
    },
}

_input_predict_schema = {
    "description": "Predict with `data`.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
            "description": "The dmatrix storing the input.",
        },
        "output_margin": {
            "type": "boolean",
            "default": False,
            "description": "Whether to output the raw untransformed margin value.",
        },
        "ntree_limit": {
            "anyOf": [{"type": "integer"}, {"enum": [None]}],
            "description": "Limit number of trees in the prediction; defaults to best_ntree_limit if defined",
        },
        "validate_features": {
            "type": "boolean",
            "default": True,
            "description": "When this is True, validate that the Booster's and data's feature_names are identical.",
        },
    },
}
_output_predict_schema = {
    "description": "Output data schema for predictions (target class labels).",
    "type": "array",
    "items": {"type": "number"},
}

_input_predict_proba_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
        }
    },
}

_output_predict_probaschema = {
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`XGBRegressor`_ gradient boosted decision trees.

.. _`XGBRegressor`: https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRegressor
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.xgboost.xgb_regressor.html",
    "import_from": "xgboost",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_schema,
        "output_predict_proba": _output_predict_schema,
    },
}

XGBRegressor: lale.operators.PlannedIndividualOp
XGBRegressor = lale.operators.make_operator(_XGBRegressorImpl, _combined_schemas)

if xgboost_installed and xgboost.__version__ >= "0.90":
    # page 58 of https://readthedocs.org/projects/xgboost/downloads/pdf/release_0.90/
    import lale.schemas

    XGBRegressor = XGBRegressor.customize_schema(
        objective=lale.schemas.JSON(
            {
                "description": "Specify the learning task and the corresponding learning objective or a custom objective function to be used.",
                "anyOf": [
                    {
                        "enum": [
                            "reg:linear",
                            "reg:logistic",
                            "reg:gamma",
                            "reg:tweedie",
                            "reg:squarederror",
                        ]
                    },
                    {"laleType": "callable"},
                ],
                "default": "reg:linear",
            }
        ),
        set_as_available=True,
    )

if xgboost_installed and xgboost.__version__ >= "1.3":
    # https://xgboost.readthedocs.io/en/latest/python/python_api.html#module-xgboost.sklearn
    XGBRegressor = XGBRegressor.customize_schema(
        monotone_constraints={
            "description": "Constraint of variable monotonicity.",
            "anyOf": [{"enum": [None]}, {"type": "string"}],
            "default": None,
        },
        interaction_constraints={
            "description": "Constraints for interaction representing permitted interactions. The constraints must be specified in the form of a nest list, e.g. [[0, 1], [2, 3, 4]], where each inner list is a group of indices of features that are allowed to interact with each other.",
            "anyOf": [{"enum": [None]}, {"type": "string"}],
            "default": None,
        },
        num_parallel_tree={
            "description": "Used for boosting random forest.",
            "anyOf": [{"enum": [None]}, {"type": "integer"}],
            "default": None,
        },
        validate_parameters={
            "description": "Give warnings for unknown parameter.",
            "anyOf": [{"enum": [None]}, {"type": "boolean"}, {"type": "integer"}],
            "default": None,
        },
        gpu_id={
            "description": "Device ordinal.",
            "anyOf": [
                {"type": "integer"},
                {"enum": [None]},
            ],
            "default": None,
        },
        max_depth={
            "description": "Maximum tree depth for base learners.",
            "anyOf": [
                {
                    "type": "integer",
                    "minimum": 0,
                    "distribution": "uniform",
                    "minimumForOptimizer": 1,
                    "maximumForOptimizer": 7,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        learning_rate={
            "description": """Boosting learning rate (xgb's "eta")""",
            "anyOf": [
                {
                    "type": "number",
                    "distribution": "loguniform",
                    "minimumForOptimizer": 0.02,
                    "maximumForOptimizer": 1,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        booster={
            "description": "Specify which booster to use.",
            "enum": ["gbtree", "gblinear", "dart", None],
            "default": None,
        },
        tree_method={
            "description": """Specify which tree method to use.
Default to auto. If this parameter is set to default, XGBoost will choose the most conservative option available.
Refer to https://xgboost.readthedocs.io/en/latest/parameter.html. """,
            "enum": ["auto", "exact", "approx", "hist", "gpu_hist", None],
            "default": None,
        },
        gamma={
            "description": "Minimum loss reduction required to make a further partition on a leaf node of the tree.",
            "anyOf": [
                {
                    "type": "number",
                    "minimum": 0,
                    "maximumForOptimizer": 1.0,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        min_child_weight={
            "description": "Minimum sum of instance weight(hessian) needed in a child.",
            "anyOf": [
                {
                    "type": "integer",
                    "distribution": "uniform",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 20,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        max_delta_step={
            "description": "Maximum delta step we allow each tree's weight estimation to be.",
            "anyOf": [{"enum": [None]}, {"type": "integer"}],
            "default": None,
        },
        subsample={
            "description": "Subsample ratio of the training instance.",
            "anyOf": [
                {
                    "type": "number",
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.01,
                    "maximumForOptimizer": 1.0,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        colsample_bytree={
            "description": "Subsample ratio of columns when constructing each tree.",
            "anyOf": [
                {
                    "type": "number",
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "maximum": 1,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.1,
                    "maximumForOptimizer": 1.0,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        colsample_bylevel={
            "description": "Subsample ratio of columns for each split, in each level.",
            "anyOf": [
                {
                    "type": "number",
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "maximum": 1,
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.1,
                    "maximumForOptimizer": 1.0,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        colsample_bynode={
            "description": "Subsample ratio of columns for each split.",
            "anyOf": [
                {
                    "type": "number",
                    "minimum": 0,
                    "exclusiveMinimum": True,
                    "maximum": 1,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        reg_alpha={
            "description": "L1 regularization term on weights",
            "anyOf": [
                {
                    "type": "number",
                    "distribution": "uniform",
                    "minimumForOptimizer": 0,
                    "maximumForOptimizer": 1,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        reg_lambda={
            "description": "L2 regularization term on weights",
            "anyOf": [
                {
                    "type": "number",
                    "distribution": "uniform",
                    "minimumForOptimizer": 0.1,
                    "maximumForOptimizer": 1,
                },
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        scale_pos_weight={
            "description": "Balancing of positive and negative weights.",
            "anyOf": [
                {"type": "number"},
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        base_score={
            "description": "The initial prediction score of all instances, global bias.",
            "anyOf": [
                {"type": "number"},
                {"enum": [None], "forOptimizer": False},
            ],
            "default": None,
        },
        missing={
            "anyOf": [
                {
                    "type": "number",
                },
                {
                    "enum": [None, np.NaN],
                },
            ],
            "default": np.NaN,
            "description": "Value in the data which needs to be present as a missing value. If"
            " If None, defaults to np.nan.",
        },
        verbosity={
            "description": "The degree of verbosity.",
            "anyOf": [{"type": "integer"}, {"enum": [None]}],
            "default": None,
            "minimum": 0,
            "maximum": 3,
        },
        set_as_available=True,
    )

if xgboost_installed and xgboost.__version__ >= "1.5":
    # https://xgboost.readthedocs.io/en/latest/python/python_api.html#module-xgboost.sklearn
    XGBRegressor = XGBRegressor.customize_schema(
        enable_categorical={
            "type": "boolean",
            "description": """Experimental support for categorical data.
Do not set to true unless you are interested in development.
Only valid when gpu_hist and dataframe are used.""",
            "default": False,
        },
        predictor={
            "anyOf": [{"type": "string"}, {"enum": [None]}],
            "description": """Force XGBoost to use specific predictor,
available choices are [cpu_predictor, gpu_predictor].""",
            "default": None,
        },
        set_as_available=True,
    )

lale.docstrings.set_docstrings(XGBRegressor)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
import sklearn.feature_extraction.text

import lale.docstrings
import lale.operators


class _TfidfVectorizerImpl:
    def __init__(self, **hyperparams):
        if "dtype" in hyperparams and hyperparams["dtype"] == "float64":
            hyperparams = {**hyperparams, "dtype": np.float64}
        self._hyperparams = hyperparams
        self._wrapped_model = sklearn.feature_extraction.text.TfidfVectorizer(
            **self._hyperparams
        )

    def fit(self, X, y=None):
        if isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame):
            X = X.squeeze().astype("U")
        self._wrapped_model.fit(X, y)
        return self

    def transform(self, X):
        if isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame):
            X = X.squeeze().astype("U")
        return self._wrapped_model.transform(X)


_hyperparams_schema = {
    "description": "Convert a collection of raw documents to a matrix of TF-IDF features.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "input",
                "encoding",
                "decode_error",
                "strip_accents",
                "lowercase",
                "preprocessor",
                "tokenizer",
                "analyzer",
                "stop_words",
                "ngram_range",
                "max_df",
                "min_df",
                "max_features",
                "vocabulary",
                "binary",
                "dtype",
                "norm",
                "use_idf",
                "smooth_idf",
                "sublinear_tf",
            ],
            "relevantToOptimizer": [
                "analyzer",
                "ngram_range",
                "max_df",
                "min_df",
                "binary",
                "norm",
                "use_idf",
                "smooth_idf",
                "sublinear_tf",
            ],
            "additionalProperties": False,
            "properties": {
                "input": {
                    "enum": ["filename", "file", "content"],
                    "default": "content",
                },
                "encoding": {"type": "string", "default": "utf-8"},
                "decode_error": {
                    "enum": ["strict", "ignore", "replace"],
                    "default": "strict",
                },
                "strip_accents": {"enum": ["ascii", "unicode", None], "default": None},
                "lowercase": {"type": "boolean", "default": True},
                "preprocessor": {
                    "anyOf": [
                        {"laleType": "callable", "forOptimizer": False},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "tokenizer": {
                    "anyOf": [
                        {"laleType": "callable", "forOptimizer": False},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "analyzer": {
                    "anyOf": [
                        {"enum": ["word", "char", "char_wb"]},
                        {"laleType": "callable", "forOptimizer": False},
                    ],
                    "default": "word",
                },
                "stop_words": {
                    "anyOf": [{"enum": [None]}, {"type": "array"}, {"type": "string"}],
                    "default": None,
                },
                "token_pattern": {"type": "string", "default": "(?u)\\b\\w\\w+\\b"},
                "ngram_range": {
                    "default": (1, 1),
                    "anyOf": [
                        {
                            "type": "array",
                            "laleType": "tuple",
                            "minItemsForOptimizer": 2,
                            "maxItemsForOptimizer": 2,
                            "items": {
                                "type": "integer",
                                "minimumForOptimizer": 1,
                                "maximumForOptimizer": 3,
                            },
                            "forOptimizer": False,
                        },
                        {"enum": [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)]},
                    ],
                },
                "max_df": {
                    "anyOf": [
                        {
                            "description": "float in range [0.0, 1.0]",
                            "type": "number",
                            "minimum": 0.0,
                            "maximum": 1.0,
                            "minimumForOptimizer": 0.8,
                            "maximumForOptimizer": 0.9,
                            "distribution": "uniform",
                        },
                        {"type": "integer", "forOptimizer": False},
                    ],
                    "default": 1.0,
                },
                "min_df": {
                    "anyOf": [
                        {
                            "description": "float in range [0.0, 1.0]",
                            "type": "number",
                            "minimum": 0.0,
                            "maximum": 1.0,
                            "minimumForOptimizer": 0.0,
                            "maximumForOptimizer": 0.1,
                            "distribution": "uniform",
                        },
                        {"type": "integer", "forOptimizer": False},
                    ],
                    "default": 1,
                },
                "max_features": {
                    "anyOf": [
                        {"type": "integer", "minimum": 1, "maximumForOptimizer": 10000},
                        {"enum": [None]},
                    ],
                    "default": None,
                },
                "vocabulary": {
                    "description": "XXX TODO XXX, Mapping or iterable, optional",
                    "anyOf": [{"type": "object"}, {"enum": [None]}],
                    "default": None,
                },
                "binary": {"type": "boolean", "default": False},
                "dtype": {
                    "description": "XXX TODO XXX, type, optional",
                    "type": "string",
                    "default": "float64",
                },
                "norm": {"enum": ["l1", "l2", None], "default": "l2"},
                "use_idf": {"type": "boolean", "default": True},
                "smooth_idf": {"type": "boolean", "default": True},
                "sublinear_tf": {"type": "boolean", "default": False},
            },
        },
        {
            "description": "tokenizer, only applies if analyzer == 'word'",
            "anyOf": [
                {"type": "object", "properties": {"analyzer": {"enum": ["word"]}}},
                {"type": "object", "properties": {"tokenizer": {"enum": [None]}}},
            ],
        },
        {
            "description": "stop_words can be a list only if analyzer == 'word'",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {"stop_words": {"not": {"type": "array"}}},
                },
                {"type": "object", "properties": {"analyzer": {"enum": ["word"]}}},
            ],
        },
    ],
}

_input_fit_schema = {
    "description": "Input data schema for training the TfidfVectorizer from scikit-learn.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "anyOf": [
                {"type": "array", "items": {"type": "string"}},
                {
                    "type": "array",
                    "items": {
                        "type": "array",
                        "minItems": 1,
                        "maxItems": 1,
                        "items": {"type": "string"},
                    },
                },
            ],
        },
        "y": {"description": "Target class labels; the array is over samples."},
    },
}

_input_transform_schema = {
    "description": "Input data schema for predictions using the TfidfVectorizer model from scikit-learn.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "anyOf": [
                {"type": "array", "items": {"type": "string"}},
                {
                    "type": "array",
                    "items": {
                        "type": "array",
                        "minItems": 1,
                        "maxItems": 1,
                        "items": {"type": "string"},
                    },
                },
            ],
        }
    },
}

_output_transform_schema = {
    "description": "Output data schema for predictions (projected data) using the TfidfVectorizer model from scikit-learn.",
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`TF-IDF vectorizer`_ transformer from scikit-learn for turning text into term frequency - inverse document frequency numeric features.

.. _`TF-IDF vectorizer`: https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.tfidf_vectorizer.html",
    "import_from": "sklearn.feature_extraction.text",
    "type": "object",
    "tags": {"pre": ["text"], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


TfidfVectorizer = lale.operators.make_operator(_TfidfVectorizerImpl, _combined_schemas)

lale.docstrings.set_docstrings(TfidfVectorizer)



from numpy import inf, nan
from sklearn.linear_model import OrthogonalMatchingPursuitCV as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _OrthogonalMatchingPursuitCVImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for OrthogonalMatchingPursuitCV    Cross-validated Orthogonal Matching Pursuit model (OMP).",
    "allOf": [
        {
            "type": "object",
            "required": [
                "copy",
                "fit_intercept",
                "normalize",
                "max_iter",
                "cv",
                "n_jobs",
                "verbose",
            ],
            "relevantToOptimizer": [
                "copy",
                "fit_intercept",
                "normalize",
                "max_iter",
                "cv",
            ],
            "additionalProperties": False,
            "properties": {
                "copy": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether the design matrix X must be copied by the algorithm",
                },
                "fit_intercept": {
                    "type": "boolean",
                    "default": True,
                    "description": "whether to calculate the intercept for this model",
                },
                "normalize": {
                    "type": "boolean",
                    "default": True,
                    "description": "This parameter is ignored when ``fit_intercept`` is set to False",
                },
                "max_iter": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimumForOptimizer": 10,
                            "maximumForOptimizer": 1000,
                            "distribution": "uniform",
                        },
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Maximum numbers of iterations to perform, therefore maximum features to include",
                },
                "cv": {
                    "description": """Cross-validation as integer or as object that has a split function.
                        The fit method performs cross validation on the input dataset for per
                        trial, and uses the mean cross validation performance for optimization.
                        This behavior is also impacted by handle_cv_failure flag.
                        If integer: number of folds in sklearn.model_selection.StratifiedKFold.
                        If object with split function: generator yielding (train, test) splits
                        as arrays of indices. Can use any of the iterators from
                        https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators.""",
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "default": 5,
                            "minimumForOptimizer": 3,
                            "maximumForOptimizer": 4,
                            "distribution": "uniform",
                        },
                        {"laleType": "Any", "forOptimizer": False},
                    ],
                },
                "n_jobs": {
                    "anyOf": [{"type": "integer"}, {"enum": [None]}],
                    "default": 1,
                    "description": "Number of CPUs to use during the cross validation",
                },
                "verbose": {
                    "anyOf": [{"type": "boolean"}, {"type": "integer"}],
                    "default": False,
                    "description": "Sets the verbosity amount",
                },
            },
        },
        {
            "XXX TODO XXX": "Parameter: copy > only helpful if x is already fortran-ordered, otherwise a copy is made anyway"
        },
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the model using X, y as training data.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training data.",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the linear model",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"laleType": "Any", "XXX TODO XXX": "item type"},
                    "XXX TODO XXX": "array_like or sparse matrix, shape (n_samples, n_features)",
                },
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
            "description": "Samples.",
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Returns predicted values.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.linear_model.OrthogonalMatchingPursuitCV#sklearn-linear_model-orthogonalmatchingpursuitcv",
    "import_from": "sklearn.linear_model",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "regressor"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
OrthogonalMatchingPursuitCV = make_operator(
    _OrthogonalMatchingPursuitCVImpl, _combined_schemas
)

set_docstrings(OrthogonalMatchingPursuitCV)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.neighbors

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Hyperparameter schema for the KNeighborsRegressor model from scikit-learn.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "n_neighbors",
                "weights",
                "algorithm",
                "leaf_size",
                "p",
                "metric",
                "metric_params",
                "n_jobs",
            ],
            "relevantToOptimizer": [
                "n_neighbors",
                "weights",
                "algorithm",
                "p",
                "metric",
            ],
            "properties": {
                "n_neighbors": {
                    "description": "Number of neighbors to use by default for kneighbors queries.",
                    "type": "integer",
                    "distribution": "uniform",
                    "minimum": 1,
                    "laleMaximum": "X/maxItems",  # number of rows
                    "default": 5,
                    "maximumForOptimizer": 100,
                },
                "weights": {
                    "description": "Weight function used in prediction.",
                    "enum": ["uniform", "distance"],
                    "default": "uniform",
                },
                "algorithm": {
                    "description": "Algorithm used to compute the nearest neighbors.",
                    "enum": ["ball_tree", "kd_tree", "brute", "auto"],
                    "default": "auto",
                },
                "leaf_size": {
                    "description": "Leaf size passed to BallTree or KDTree.",
                    "type": "integer",
                    "distribution": "uniform",
                    "minimum": 1,
                    "default": 30,
                    "maximumForOptimizer": 100,
                },
                "p": {
                    "description": "Power parameter for the Minkowski metric.",
                    "type": "integer",
                    "distribution": "uniform",
                    "minimum": 1,
                    "default": 2,
                    "maximumForOptimizer": 3,
                },
                "metric": {
                    "description": "The distance metric to use for the tree.",
                    "enum": ["euclidean", "manhattan", "minkowski"],
                    "default": "minkowski",
                },
                "metric_params": {
                    "description": "Additional keyword arguments for the metric function.",
                    "anyOf": [
                        {"enum": [None]},
                        {
                            "type": "object",
                            "propertyNames": {"pattern": "[_a-zA-Z][_a-zA-Z0-9]*"},
                        },
                    ],
                    "default": None,
                },
                "n_jobs": {
                    "description": "Number of parallel jobs to run for the neighbor search.",
                    "anyOf": [
                        {
                            "description": "1 unless in joblib.parallel_backend context.",
                            "enum": [None],
                        },
                        {"description": "Use all processors.", "enum": [(-1)]},
                        {
                            "description": "Number of CPU cores.",
                            "type": "integer",
                            "minimum": 1,
                        },
                    ],
                    "default": None,
                },
            },
        },
    ],
}

_input_fit_schema = {
    "description": "Input data schema for training the KNeighborsRegressor model from scikit-learn.",
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number"},
            },
        },
        "y": {
            "description": "Target class labels; the array is over samples.",
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
            ],
        },
    },
}

_input_predict_schema = {
    "description": "Input data schema for predictions using the KNeighborsRegressor model from scikit-learn.",
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        }
    },
}

_output_predict_schema = {
    "description": "Returns predicted values.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "array", "items": {"type": "number"}}},
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`K nearest neighbors regressor`_ from scikit-learn.

.. _`K nearest neighbors regressor`: https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsRegressor.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.k_neighbors_regressor.html",
    "import_from": "sklearn.neighbors",
    "type": "object",
    "tags": {
        "pre": ["~categoricals"],
        "op": ["estimator", "regressor", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}

KNeighborsRegressor = lale.operators.make_operator(
    sklearn.neighbors.KNeighborsRegressor, _combined_schemas
)

lale.docstrings.set_docstrings(KNeighborsRegressor)
